# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import os
from typing import Optional, Union

from geti_sdk.demos.constants import DEFAULT_DATA_PATH

from .download_helpers import download_file, validate_hash

VIDEO_PERSON_CAR_BIKE_PATH = os.path.join(
    DEFAULT_DATA_PATH, "person-bicycle-car-detection.mp4"
)


def get_person_car_bike_video(
    video_path: Optional[Union[str, os.PathLike]] = None
) -> str:
    """
    Get the path to the 'person-bicycle-car-detection.mp4' video file that is used for
    the notebook demos in the Geti SDK

    :param video_path: Optional file path to video. Only specify this if you have already
        downloaded the video to a different path than the default data path in the SDK.
    :return: Path to the video
    """
    if video_path is None:
        video_path = VIDEO_PERSON_CAR_BIKE_PATH
    if os.path.isfile(video_path):
        video_file_path = video_path
    else:
        video_file_path = download_file(
            url="https://storage.openvinotoolkit.org/data/test_data/videos/person-bicycle-car-detection.mp4",
            target_folder=os.path.dirname(video_path),
        )

    # Compare hashes
    expected_hash = "7eafb94b3491e7554d92637596ddaaf1c69fdb8421c3d49eb09df4b7d05788c76391cb791cc2c8197e66d352306328d09c8d72e42733b1f1a311cd1e35ee1cce"
    validate_hash(file_path=video_file_path, expected_hash=expected_hash)
    return video_file_path



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
from collections import UserList
from typing import Any, Dict, List, Optional, Sequence

from geti_sdk.data_models.algorithms import Algorithm
from geti_sdk.data_models.enums import TaskType

DEFAULT_ALGORITHMS = {
    "classification": "Custom_Image_Classification_EfficinetNet-B0",
    "detection": "Custom_Object_Detection_Gen3_ATSS",
    "segmentation": "Custom_Semantic_Segmentation_Lite-HRNet-18-mod2_OCR",
    "anomaly_classification": "ote_anomaly_classification_padim",
    "anomaly_detection": "ote_anomaly_classification_padim",
    "anomaly_segmentation": "ote_anomaly_segmentation_padim",
    "rotated_detection": "Custom_Rotated_Detection_via_Instance_Segmentation_MaskRCNN_ResNet50",
    "instance_segmentation": "Custom_Counting_Instance_Segmentation_MaskRCNN_ResNet50",
}


class AlgorithmList(UserList):
    """
    A list containing the algorithms supported in Intel® Geti™.
    """

    def __init__(self, data: Optional[Sequence[Algorithm]] = None):
        self.data: List[Algorithm] = []
        if data is not None:
            super().__init__(list(data))

    @staticmethod
    def from_rest(rest_input: Dict[str, Any]) -> "AlgorithmList":
        """
        Create an AlgorithmList from the response of the /supported_algorithms REST
        endpoint in Intel® Geti™.

        :param rest_input: Dictionary retrieved from the /supported_algorithms REST
            endpoint
        :return: AlgorithmList holding the information related to the supported
            algorithms in Intel® Geti™
        """
        algorithm_list = AlgorithmList([])
        if "items" in rest_input:
            algo_rest = rest_input["items"]
        elif "supported_algorithms" in rest_input:
            algo_rest = rest_input["supported_algorithms"]
        else:
            raise KeyError(
                "The input dictionary does not contain the supported algorithms."
            )
        algo_rest_list = copy.deepcopy(algo_rest)
        for algorithm_dict in algo_rest_list:
            algorithm_list.append(Algorithm(**algorithm_dict))
        algorithm_list.sort(key=lambda x: x.gigaflops)
        return algorithm_list

    def get_by_model_template(self, model_template_id: str) -> Algorithm:
        """
        Retrieve an algorithm from the list by its model_template_id.

        :param model_template_id: Name of the model template to get the Algorithm
            information for
        :return: Algorithm holding the algorithm details
        """
        for algo in self.data:
            if algo.model_template_id == model_template_id:
                return algo
        raise ValueError(
            f"Algorithm for model template {model_template_id} was not found in the "
            f"list of supported algorithms."
        )

    def get_by_task_type(self, task_type: TaskType) -> "AlgorithmList":
        """
        Return a list of supported algorithms for a particular task type.

        :param task_type: TaskType to get the supported algorithms for
        :return: List of supported algorithms for the task type
        """
        return AlgorithmList(
            [algo for algo in self.data if algo.task_type == task_type]
        )

    @property
    def summary(self) -> str:
        """
        Return a string that gives a very brief summary of the algorithm list.

        :return: String holding a brief summary of the list of algorithms
        """
        summary_str = "Algorithms:\n"
        for algorithm in self.data:
            summary_str += (
                f"  Name: {algorithm.name}\n"
                f"    Task type: {algorithm.task_type}\n"
                f"    Model size: {algorithm.model_size}\n"
                f"    Gigaflops: {algorithm.gigaflops}\n"
                f"    Recommended for: {algorithm.performance_category}\n\n"
            )
        return summary_str

    def get_by_name(self, name: str) -> Algorithm:
        """
        Retrieve an algorithm from the list by its algorithm_name.

        :param name: Name of the Algorithm to get
        :return: Algorithm holding the algorithm details
        """
        for algo in self.data:
            if algo.name == name:
                return algo
        raise ValueError(
            f"Algorithm named {name} was not found in the "
            f"list of supported algorithms."
        )

    def get_default_for_task_type(self, task_type: TaskType) -> Algorithm:
        """
        Return the default algorithm for a given task type. If there is no algorithm
        for the task type in the AlgorithmList, this method will raise a ValueError.

        :param task_type: TaskType of the task to get the default algorithm for
        :raises: ValueError if there are no available algorithms for the specified
            task_type in the AlgorithmList
        :return: Default algorithm for the task
        """
        task_algos = self.get_by_task_type(task_type=task_type)
        default = [algo for algo in task_algos if algo.default_algorithm]
        if len(default) == 1:
            return default[0]
        else:
            # The old method used in Geti v1.8 and lower. Keep for backwards
            # compatibility
            return self.get_by_model_template(DEFAULT_ALGORITHMS[str(task_type)])



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from enum import Enum
from typing import List


class JobState(Enum):
    """
    Enum representing the state of a job on the Intel® Geti™ server.
    """

    IDLE = "idle"
    RUNNING = "running"
    PAUSED = "paused"
    FINISHED = "finished"
    ERROR = "error"
    FAILED = "failed"
    CANCELLED = "cancelled"
    INACTIVE = "inactive"
    SCHEDULED = "scheduled"

    def __str__(self) -> str:
        """
        Return the string representation of the JobState instance.
        """
        return self.value

    @classmethod
    def active_states(cls) -> List["JobState"]:
        """
        Return a list of JobState instance which represent jobs that are still active.

        :return: List of JobState instances
        """
        return [cls.IDLE, cls.PAUSED, cls.RUNNING]

    @classmethod
    def inactive_states(cls) -> List["JobState"]:
        """
        Return a list of JobState instance which represent jobs that are inactive,
        i.e. cancelled, errored or finished successfully.

        :return: List of JobState instances
        """
        return [cls.FINISHED, cls.ERROR, cls.FAILED, cls.FINISHED, cls.INACTIVE]



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
from typing import Any, Dict, List

import attr

from geti_sdk.data_models import Annotation, Prediction
from geti_sdk.data_models.predictions import ResultMedium
from geti_sdk.data_models.utils import attr_value_serializer, remove_null_fields
from geti_sdk.rest_converters import AnnotationRESTConverter


class PredictionRESTConverter:
    """
    Class to convert REST representations of predictions into Prediction entities.
    """

    @staticmethod
    def from_dict(prediction: Dict[str, Any]) -> Prediction:
        """
        Create a Prediction object from a dictionary returned by the
        /predictions REST endpoint in the Intel® Geti™ platform.

        :param prediction: dictionary representing a Prediction, which
            contains all prediction annotations for a certain media entity
        :return: Prediction object
        """
        input_copy = copy.deepcopy(prediction)
        annotations: List[Annotation] = []
        prediction_dicts = input_copy.pop("predictions", None)
        if prediction_dicts is None:
            # Geti versions lower than 1.13 still use 'annotations' as key
            prediction_dicts = prediction.get("annotations")
        for annotation in prediction_dicts:
            if not isinstance(annotation, Annotation):
                annotations.append(
                    AnnotationRESTConverter.annotation_from_dict(annotation)
                )
            else:
                annotations.append(annotation)
        media_identifier_dict = prediction.get("media_identifier", None)
        if media_identifier_dict is not None:
            media_identifier = AnnotationRESTConverter._media_identifier_from_dict(
                prediction["media_identifier"]
            )
        else:
            media_identifier = None

        result_media: List[ResultMedium] = []
        for result_medium in prediction.get("maps", []):
            if not isinstance(result_medium, ResultMedium):
                result_media.append(ResultMedium(**result_medium))
            else:
                result_media.append(result_medium)
        input_copy.update(
            {
                "annotations": annotations,
                "media_identifier": media_identifier,
                "maps": result_media,
            }
        )
        return Prediction(**input_copy)

    @staticmethod
    def to_dict(prediction: Prediction, deidentify: bool = True) -> Dict[str, Any]:
        """
        Convert a Prediction to a dictionary. By default, removes any ID
        fields in the output dictionary

        :param prediction: Prediction object to convert
        :param deidentify: True to remove any unique database ID fields in the output,
            False to keep these fields. Defaults to True
        :return: Dictionary holding the serialized Prediction data
        """
        if deidentify:
            prediction.deidentify()
        prediction_dict = attr.asdict(
            prediction, recurse=True, value_serializer=attr_value_serializer
        )
        remove_null_fields(prediction_dict)
        return prediction_dict



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import glob
import json
import logging
import os
import warnings
from random import sample
from typing import Any, Dict, List, Optional, Union

from geti_sdk.data_models import Annotation, TaskType
from geti_sdk.data_models.media import MediaInformation
from geti_sdk.rest_converters import AnnotationRESTConverter
from geti_sdk.rest_converters.annotation_rest_converter import (
    NormalizedAnnotationRESTConverter,
)

from .base_annotation_reader import AnnotationReader


class GetiAnnotationReader(AnnotationReader):
    """
    AnnotationReader for loading annotation files in Intel® Geti™ format.
    """

    def __init__(
        self,
        base_data_folder: str,
        annotation_format: str = ".json",
        task_type: Optional[Union[TaskType, str]] = None,
        label_names_to_include: Optional[List[str]] = None,
    ):
        """
        :param base_data_folder: Path to the folder containing the annotations
        :param annotation_format: Extension of the annotation files. Defaults to ".json"
        :param task_type: Optional task type to prepare the annotations for. Can also
            be specified later.
        :param label_names_to_include: Names of the labels that should be included
            when reading annotation data. This can be used to filter the annotations
            for certain labels.
        """
        if annotation_format != ".json":
            raise ValueError(
                f"Annotation format {annotation_format} is currently not"
                f" supported by the GetiAnnotationReader"
            )
        super().__init__(
            base_data_folder=base_data_folder,
            annotation_format=annotation_format,
            task_type=task_type,
        )
        self._label_names_to_include = label_names_to_include
        self._normalized_annotations = self._has_normalized_annotations()

    def _get_label_names(self, all_labels: List[str]) -> List[str]:
        """
        Return the labels for the task type the annotation reader is currently set to.

        :param all_labels: List of all label names in the project
        """
        if self._label_names_to_include is None:
            logging.info("No label mapping defined, including all labels")
            labels = all_labels
        else:
            labels = [
                label for label in all_labels if label in self._label_names_to_include
            ]
        return labels

    def _get_raw_annotation_data(self, filename: str) -> Dict[str, Any]:
        """
        Read the annotation data from the file at `filename`

        :param filename: Name of the annotation file to read
        :return: Dictionary holding the annotation data
        """
        filepath = glob.glob(
            os.path.join(self.base_folder, f"{filename}{self.annotation_format}"),
        )
        if len(filepath) > 1:
            warnings.warn(
                f"Multiple matching annotation files found for image with "
                f"name {filename}. Skipping this image..."
            )
            data = {"annotations": []}
        elif len(filepath) == 0:
            logging.info(
                f"No matching annotation file found for image with name {filename}."
                f" Skipping this image..."
            )
            data = {"annotations": []}
        else:
            filepath = filepath[0]
            with open(filepath, "r") as f:
                data = json.load(f)
        return data

    def get_data(
        self,
        filename: str,
        label_name_to_id_mapping: dict,
        media_information: MediaInformation,
        preserve_shape_for_global_labels: bool = False,
    ) -> List[Annotation]:
        """
        Return the annotation data for the dataset item corresponding to `filename`.

        :param filename: name of the item to get the annotation data for.
        :param label_name_to_id_mapping: mapping of label name to label id.
        :param media_information: MediaInformation object containing information
            (e.g. width, height) about the media item to upload the annotation for
        :param preserve_shape_for_global_labels: False to convert shapes for global
            tasks to full rectangles (required for classification like tasks in
            Intel® Geti™ projects), True to preserve such shapes. This parameter
            should be:

             - False when uploading annotations to a single task project
             - True when uploading annotations for a classification like task,
                following a local task in a task chain project.

        :return: List of Annotation objects containing all annotations for the given
            dataset item.
        """
        data = self._get_raw_annotation_data(filename=filename)

        new_annotations = []
        for annotation in data["annotations"]:
            if self._normalized_annotations:
                annotation_object = (
                    NormalizedAnnotationRESTConverter.normalized_annotation_from_dict(
                        annotation,
                        image_width=media_information.width,
                        image_height=media_information.height,
                    )
                )
            else:
                annotation_object = AnnotationRESTConverter.annotation_from_dict(
                    annotation
                )
            for label in annotation_object.labels:
                label.id = label_name_to_id_mapping[label.name]
            for label_dict in annotation["labels"]:
                if self.task_type is not None:
                    if label_dict["name"] not in self._get_label_names(
                        list(label_name_to_id_mapping.keys())
                    ):
                        annotation_object.pop_label_by_name(
                            label_name=label_dict["name"]
                        )
            new_annotations.append(annotation_object)
        return new_annotations

    def get_all_label_names(self) -> List[str]:
        """
        Retrieve the unique label names for all annotations in the annotation folder

        :return: List of label names
        """
        logging.info(f"Reading annotation files in folder {self.base_folder}...")
        unique_label_names = []
        annotation_files = glob.glob(
            os.path.join(self.base_folder, f"*{self.annotation_format}"),
        )
        if len(annotation_files) == 0:
            raise ValueError(
                f"No valid annotation files were found in folder {self.base_folder}"
            )
        for annotation_file in annotation_files:
            with open(annotation_file, "r") as f:
                data = json.load(f)
            annotations = data.get("annotations", None)
            if annotations is None:
                raise ValueError(
                    f"Annotation file '{annotation_file}' does not contain any "
                    f"annotations. Please make sure that this is a valid "
                    f"annotation file."
                )
            for annotation in annotations:
                labels = [label["name"] for label in annotation["labels"]]
                for label in labels:
                    if label not in unique_label_names:
                        unique_label_names.append(label)
        return unique_label_names

    def _has_normalized_annotations(self) -> bool:
        """
        Check if the annotation files belonging to this annotation reader are normalized
        """
        filenames = self.get_data_filenames()
        n_sample = min(len(filenames), 50)
        if n_sample == 50:
            list_to_check = sample(filenames, n_sample)
        else:
            list_to_check = filenames

        NORMALIZED_KEY = "normalized"
        PIXEL_KEY = "pixel"

        annotation_stats = {NORMALIZED_KEY: 0, PIXEL_KEY: 0}
        for filename in list_to_check:
            data = self._get_raw_annotation_data(filename=filename)
            for annotation_dict in data["annotations"]:
                annotation_object = AnnotationRESTConverter.annotation_from_dict(
                    annotation_dict
                )
                shape = annotation_object.shape
                x_max, y_max = shape.x_max, shape.y_max
                if x_max <= 1 and y_max <= 1:
                    annotation_stats[NORMALIZED_KEY] += 1
                else:
                    annotation_stats[PIXEL_KEY] += 1

        if annotation_stats[NORMALIZED_KEY] == 0:
            return False
        elif annotation_stats[PIXEL_KEY] == 0:
            logging.info(
                "Legacy annotation format detected. The annotations you are trying to "
                "upload were most likely downloaded from a pre-production version of "
                "the Intel Geti software. They will be converted to the latest "
                "annotation format upon upload to the Intel Geti platform. "
            )
            return True
        else:
            raise ValueError(
                f"The annotation directory '{self.base_folder}' contains both "
                f"normalized ({annotation_stats[NORMALIZED_KEY]} shapes) and "
                f"non-normalized ({annotation_stats[PIXEL_KEY]} shapes) objects. "
                f"Unable to parse annotation data."
            )



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
from typing import Any, Dict

from geti_sdk.data_models import Project
from geti_sdk.data_models.utils import remove_null_fields
from geti_sdk.utils import deserialize_dictionary


class ProjectRESTConverter:
    """
    Class that handles conversion of Intel® Geti™ REST output for project entities to
    objects and vice versa.
    """

    @classmethod
    def from_dict(cls, project_input: Dict[str, Any]) -> Project:
        """
        Create a Project from a dictionary representing a project, as
        returned by the /projects endpoint on the Intel® Geti™ platform.

        :param project_input: Dictionary representing a project, as returned by the
            Intel® Geti™ server
        :return: Project object representing the project given in `project_input`
        """
        prepared_project = copy.deepcopy(project_input)
        for connection in prepared_project["pipeline"]["connections"]:
            from_ = connection.pop("from", None)
            if from_ is not None:
                connection.update({"from_": from_})
        return deserialize_dictionary(prepared_project, output_type=Project)

    @classmethod
    def to_dict(cls, project: Project) -> Dict[str, Any]:
        """
        Convert the `project` to its dictionary representation.
        This functions removes database UID's and optional fields that are `None`
        from the output dictionary, to make the output more compact and improve
        readability.

        :param project: Project to convert to dictionary
        :return:
        """
        project.deidentify()
        project_data = project.to_dict()
        remove_null_fields(project_data)
        return project_data



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
from typing import Any, Dict, List, Optional

import attr

from geti_sdk.data_models.performance import Performance


@attr.define(slots=False)
class StatusSummary:
    """
    Summary of the status of a project or task on the GETi cluster.

    NOTE: the 'message' attribute was removed in Geti 1.1

    :var progress: Training progress, if a model is being trained
    :var message: Optional Human readable message describing the status
    """

    progress: float
    message: Optional[str] = None

    def __attrs_post_init__(self):
        """
        Initialize private attributes
        """
        self._user_friendly_message = "Unknown status"

    @classmethod
    def from_dict(cls, status_dict: Dict[str, Any]) -> "StatusSummary":
        """
        Create a StatusSummary object from a dictionary.

        :param status_dict: Dictionary representing a status, as returned by the GETi
            /status and /jobs endpoints
        :return: StatusSummary object holding the status data contained in `status_dict`
        """
        return cls(**status_dict)

    @property
    def user_friendly_message(self) -> str:
        """
        Return a message describing the status in a human readable format
        """
        if self.message is not None:
            return self.message
        else:
            return self._user_friendly_message

    @user_friendly_message.setter
    def user_friendly_message(self, message: str):
        """
        Set the user friendly message to describe the status
        """
        self._user_friendly_message = message


@attr.define
class LabelAnnotationRequirements:
    """
    Detailed information regarding the required number of annotations for a
    specific label.

    :var id: Unique database ID of the label
    :var label_name: Name of the label
    :var label_color: Color of the label
    :var value: Required number of annotations for this label
    """

    id: str
    label_name: str
    label_color: str
    value: int


@attr.define
class AnnotationRequirements:
    """
    Container holding the required number of annotations before auto-training can be
    triggered for a task in GETi.

    :var value: Total number of required annotations for the task
    :var details: Required annotations per label
    """

    details: List[LabelAnnotationRequirements]
    value: int


@attr.define
class TaskStatus:
    """
    Status of a single task in GETi.

    :var id: Unique database ID of the task
    :var is_training: True if a training job is currently running for the task
    :var required_annotations: AnnotationRequirements object that holds details
        related to the required number of annotations before auto-training will be
        started for the task
    :var status: StatusSummary object that contains (among others) a human readable
        message describing the status of the task
    :var title: Title of the taks
    :var n_new_annotations: Number of new annotations that have been made for this
        task since its last training round. Only used in Geti v1.1 and up
    """

    id: str
    is_training: bool
    required_annotations: AnnotationRequirements
    status: StatusSummary
    title: str
    n_new_annotations: Optional[int] = None  # Added in Geti v1.1
    ready_to_train: Optional[bool] = None  # Added in Geti v1.4

    def __attrs_post_init__(self):
        """
        Make sure task status message is set correctly
        """
        if self.is_training:
            self.status.user_friendly_message = "Training"
        elif self.required_annotations.value != 0:
            self.status.user_friendly_message = "Waiting for user annotations"


@attr.define
class ProjectStatus:
    """
    Status of a project in GETi.

    :param is_training: True if a training job is currently running for any of the
        tasks in the project
    :var n_required_annotations: Total number of required annotations for the project,
        before auto-training can be started
    :var status: StatusSummary object that contains (among others) a human readable
        message describing the status of the project
    :var tasks: List of TaskStatus objects, detailing the status of each task in the
        project
    :var n_new_annotations: Only used in Geti v1.1
    """

    is_training: bool
    n_required_annotations: int
    status: StatusSummary
    tasks: List[TaskStatus]
    n_new_annotations: Optional[int] = None  # Added in Geti v1.1
    project_performance: Optional[Performance] = None
    n_running_jobs: Optional[int] = None
    n_running_jobs_project: Optional[int] = None

    def __attrs_post_init__(self):
        """
        Make sure task status message is set correctly
        """
        if self.is_training:
            self.status.user_friendly_message = "Training"
        elif self.n_required_annotations != 0:
            self.status.user_friendly_message = "Waiting for user annotations"

    @property
    def summary(self) -> str:
        """
        Return a string that gives a very brief summary of the project status.

        :return: String holding a brief summary of the project status
        """
        summary_str = f"Project status:\n  {self.status.user_friendly_message}\n"
        for task in self.tasks:
            summary_str += (
                f"    Task: {task.title}\n"
                f"      State: {task.status.user_friendly_message}\n"
            )
            if task.is_training or task.status.progress != -1.0:
                summary_str += f"      Progress: {task.status.progress:.1f}%\n"
        if self.project_performance.score is not None:
            summary_str += (
                f"  Latest score: {self.project_performance.score*100:.1f}%\n"
            )
        return summary_str



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Any, Dict

from geti_sdk.data_models import ProjectStatus
from geti_sdk.utils import deserialize_dictionary


class StatusRESTConverter:
    """
    Class that handles conversion of Intel® Geti™ REST output for status entities to
    objects
    """

    @staticmethod
    def from_dict(project_status_dict: Dict[str, Any]) -> ProjectStatus:
        """
        Create a ProjectStatus instance from the input dictionary passed in
        `project_status_dict`.

        :param project_status_dict: Dictionary representing the status of a project on
            the Intel® Geti™ server
        :return: ProjectStatus instance, holding the status data contained in
            project_status_dict
        """
        return deserialize_dictionary(project_status_dict, output_type=ProjectStatus)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import json
import logging
import os
import shutil
import tempfile
import time
import zipfile
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import attr
import numpy as np
from model_api.adapters import OpenvinoAdapter, OVMSAdapter
from model_api.models import Model as model_api_Model
from model_api.tilers import DetectionTiler, InstanceSegmentationTiler, Tiler
from openvino.runtime import Core
from packaging.version import Version

from geti_sdk.data_models import OptimizedModel, Project, TaskConfiguration
from geti_sdk.data_models.enums.domain import Domain
from geti_sdk.data_models.label import Label
from geti_sdk.data_models.label_group import LabelGroup, LabelGroupType
from geti_sdk.data_models.label_schema import LabelSchema
from geti_sdk.data_models.predictions import Prediction, ResultMedium
from geti_sdk.deployment.predictions_postprocessing.results_converter.results_to_prediction_converter import (
    ConverterFactory,
    InferenceResultsToPredictionConverter,
)
from geti_sdk.http_session import GetiSession
from geti_sdk.rest_converters import ConfigurationRESTConverter, ModelRESTConverter

from .utils import (
    generate_ovms_model_address,
    generate_ovms_model_name,
    get_package_version_from_requirements,
    rgb_to_hex,
    target_device_is_ovms,
)

MODEL_DIR_NAME = "model"
PYTHON_DIR_NAME = "python"
REQUIREMENTS_FILE_NAME = "requirements.txt"

LABELS_CONFIG_KEY = "labels"
LABEL_TREE_KEY = "label_tree"
LABEL_GROUPS_KEY = "label_groups"
ALL_LABELS_KEY = "all_labels"

SALIENCY_KEY = "saliency_map"
ANOMALY_SALIENCY_KEY = "anomaly_map"
SEGMENTATION_SALIENCY_KEY = "soft_prediction"
FEATURE_VECTOR_KEY = "feature_vector"

TILER_MAPPING = {
    Domain.DETECTION: DetectionTiler,
    Domain.INSTANCE_SEGMENTATION: InstanceSegmentationTiler,
}

OVMS_TIMEOUT = 10  # Max time to wait for OVMS models to become available


@attr.define
class DeployedModel(OptimizedModel):
    """
    Representation of an Intel® Geti™ model that has been deployed for inference. It
    can be loaded onto a device to generate predictions.
    """

    hyper_parameters: Optional[TaskConfiguration] = attr.field(
        kw_only=True, repr=False, default=None
    )

    def __attrs_post_init__(self):
        """
        Initialize private attributes
        """
        super().__attrs_post_init__()
        self._domain: Optional[Domain] = None
        self._model_data_path: Optional[str] = None
        self._model_python_path: Optional[str] = None
        self._needs_tempdir_deletion: bool = False
        self._tempdir_path: Optional[str] = None
        self._label_schema: Optional[LabelSchema] = None

        # Attributes related to model explainability
        self._saliency_key: Optional[str] = None
        self._saliency_location: Optional[str] = None
        self._feature_vector_key: Optional[str] = None
        self._feature_vector_location: Optional[str] = None

        self._converter: Optional[InferenceResultsToPredictionConverter] = None
        self._async_callback_defined: bool = False
        self._tiling_enabled: bool = False
        self._tiler: Optional[Tiler] = None

    @property
    def model_data_path(self) -> str:
        """
        Return the path to the raw model data

        :return: path to the directory containing the raw model data
        """
        if self._model_data_path is None:
            raise ValueError(
                "Model data path has not been set yet, location of binary model data "
                "is unknown."
            )
        return self._model_data_path

    def get_data(self, source: Union[str, os.PathLike, GetiSession]):
        """
        Load the model weights from a data source. The `source` can be one of the
        following:

          1. The Intel® Geti™ platform (if an GetiSession instance is passed). In this
            case the weights will be downloaded, and extracted to a temporary directory
          2. A zip file on local disk, in this case the weights will be extracted to a
             temporary directory
          3. A folder on local disk containing the .xml and .bin file for the model

        :param source: Data source to load the weights from
        """
        if isinstance(source, (os.PathLike, str)):
            if os.path.isfile(source) and os.path.splitext(source)[1] == ".zip":
                # Extract zipfile into temporary directory
                if self._model_data_path is None:
                    temp_dir = tempfile.mkdtemp()
                    self._needs_tempdir_deletion = True
                    self._tempdir_path = temp_dir
                else:
                    temp_dir = self._model_data_path

                with zipfile.ZipFile(source, "r") as zipped_source_model:
                    zipped_source_model.extractall(temp_dir)

                # _model_data_path contains the model structure and weights
                # _model_python_path contains the custom model wrappers
                self._model_data_path = os.path.join(temp_dir, MODEL_DIR_NAME)
                self._model_python_path = os.path.join(temp_dir, PYTHON_DIR_NAME)

                self.get_data(temp_dir)

            elif os.path.isdir(source):
                source_contents = os.listdir(source)
                if MODEL_DIR_NAME in source_contents:
                    model_dir = os.path.join(source, MODEL_DIR_NAME)
                else:
                    model_dir = source
                model_dir_contents = os.listdir(model_dir)
                if (
                    "model.bin" in model_dir_contents
                    and "model.xml" in model_dir_contents
                ):
                    self._model_data_path = model_dir
                else:
                    raise ValueError(
                        f"Unable to load model data from path '{model_dir}'. Model "
                        f"file 'model.xml' and weights file 'model.bin' were not found "
                        f"at the path specified. "
                    )

                self._model_python_path = os.path.join(source, PYTHON_DIR_NAME)

            # A model is being loaded from disk, check if it is a legacy model
            # We support OTX models starting from version 1.5.0
            otx_version = get_package_version_from_requirements(
                requirements_path=os.path.join(
                    self._model_python_path, REQUIREMENTS_FILE_NAME
                ),
                package_name="otx",
            )
            if otx_version:  # Empty string if package not found
                if Version(otx_version) < Version("1.5.0"):
                    raise ValueError(
                        "\n"
                        "This deployment model is not compatible with the current SDK. Proposed solutions:\n"
                        "1. Please deploy a model using GETi Platform version 2.0.0 or higher.\n"
                        "2. Downgrade to a compatible GETi-SDK version to continue using this model.\n\n"
                    )

        elif isinstance(source, GetiSession):
            if self.base_url is None:
                raise ValueError(
                    f"Insufficient data to retrieve data for model {self}. Please set "
                    f"a base_url for the model first."
                )
            response = source.get_rest_response(
                url=self.base_url + "/export", method="GET", contenttype="zip"
            )
            filename = f"{self.name}_{self.optimization_type}_optimized.zip"
            model_dir = tempfile.mkdtemp()
            model_filepath = os.path.join(model_dir, filename)
            with open(model_filepath, "wb") as f:
                f.write(response.content)
            self._model_data_path = model_dir
            self._needs_tempdir_deletion = True
            self._tempdir_path = model_dir
            self.get_data(source=model_filepath)

    def __del__(self):
        """
        Clean up the temporary directory created to store the model data (if any). This
        method is called when the OptimizedModel object is deleted.
        """
        if self._needs_tempdir_deletion:
            if self._tempdir_path is not None and os.path.exists(self._tempdir_path):
                shutil.rmtree(self._tempdir_path)

    def load_inference_model(
        self,
        device: str = "CPU",
        configuration: Optional[Dict[str, Any]] = None,
        project: Optional[Project] = None,
        plugin_configuration: Optional[Dict[str, str]] = None,
        max_async_infer_requests: int = 0,
    ) -> None:
        """
        Load the actual model weights to a specified device.

        :param device: Device (CPU or GPU) to load the model to. Defaults to 'CPU'
        :param configuration: Optional dictionary holding additional configuration
            parameters for the model
        :param project: Optional project to which the model belongs.
            This is only used when the model is run on OVMS, in that case the
            project is needed to identify the correct model
        :param plugin_configuration: Configuration for the OpenVINO execution mode
            and plugins. This can include for example specific performance hints. For
            further details, refer to the OpenVINO documentation here:
            https://docs.openvino.ai/2022.3/openvino_docs_OV_UG_Performance_Hints.html#doxid-openvino-docs-o-v-u-g-performance-hints
        :param max_async_infer_requests: Maximum number of asynchronous infer request
            that can be processed in parallel. This depends on the properties of the
            target device. If left to 0 (the default), the optimal number of requests
            will be selected automatically.
        :return: OpenVino inference engine model that can be used to make predictions
            on images
        """
        core = Core()
        if not target_device_is_ovms(device=device):
            # Run the model locally
            model_adapter = OpenvinoAdapter(
                core,
                model=os.path.join(self.model_data_path, "model.xml"),
                weights_path=os.path.join(self.model_data_path, "model.bin"),
                device=device,
                plugin_config=plugin_configuration,
                max_num_requests=max_async_infer_requests,
            )
            if max_async_infer_requests == 0:
                # Compile model to query optimal infer requests for the device
                compiled_model = core.compile_model(model_adapter.get_model(), device)
                optimal_requests = compiled_model.get_property(
                    "OPTIMAL_NUMBER_OF_INFER_REQUESTS"
                )
                model_adapter.max_num_requests = optimal_requests
                compiled_model = None
                logging.info(
                    f"Model `{self.name}` -- Optimal number of infer "
                    f"requests: {optimal_requests}"
                )
        else:
            # Connect to an OpenVINO model server instance
            model_name = generate_ovms_model_name(project=project, model=self)
            model_address = generate_ovms_model_address(
                ovms_address=device, model_name=model_name
            )

            ovms_connected = False
            ovms_error: Optional[BaseException] = None
            t_start = time.time()
            while not ovms_connected and time.time() - t_start < OVMS_TIMEOUT:
                # If OVMS has just started, model needs some time to initialize
                try:
                    model_adapter = OVMSAdapter(model_address)
                    ovms_connected = True
                except RuntimeError as error:
                    time.sleep(0.5)
                    ovms_error = error
            if not ovms_connected:
                if ovms_error is not None:
                    raise RuntimeError("Unable to connect to OVMS") from ovms_error
                else:
                    raise RuntimeError(
                        "Unknown error encountered while connecting to OVMS"
                    )

        # Load model configuration
        config_path = os.path.join(self.model_data_path, "config.json")
        if not os.path.isfile(config_path):
            raise ValueError(
                f"Missing configuration file `config.json` for deployed model `{self}`,"
                f" unable to load inference model."
            )
        with open(config_path, "r") as config_file:
            configuration_json = json.load(config_file)
        model_type = configuration_json.get("type_of_model")

        # Update model parameters
        parameters = configuration_json.get("model_parameters")
        if configuration is not None:
            configuration.update(parameters)
        else:
            configuration = parameters

        # Parse label schema
        label_dictionary = configuration_json.get("model_parameters").pop(
            LABELS_CONFIG_KEY, None
        )
        self._parse_label_schema_from_dict(label_dictionary)

        # Load a Results-to-Prediction converter
        self._converter = ConverterFactory.create_converter(
            self.label_schema, configuration
        )
        self._domain = ConverterFactory._get_labels_domain(self.label_schema)

        model = model_api_Model.create_model(
            model=model_adapter,
            model_type=model_type,
            preload=False,
            core=core,
            configuration=configuration,
        )

        self._inference_model = model

        # Extract tiling parameters, if applicable
        tiling_parameters = configuration_json.get("tiling_parameters", None)

        enable_tiling = False
        if tiling_parameters is not None:
            enable_tiling = tiling_parameters.get("enable_tiling", False)
            if isinstance(enable_tiling, dict):
                enable_tiling = enable_tiling.get("value", False)

        if enable_tiling:
            logging.info("Tiling is enabled for this model, initializing Tiler")
            tiler_type = TILER_MAPPING.get(self._domain, None)
            if tiler_type is None:
                raise ValueError(f"Tiling is not supported for domain {self._domain}")
            # InstanceSegmentationTiler supports a `tile_classifier` model, which is
            # used to filter tiles based on their objectness score. The tile
            # classifier model will be exported to the same directory as the instance
            # segmentation model. If it's there, we will load it and add it to the Tiler
            classifier_name = "tile_classifier"
            tiler_arguments = {"model": model, "execution_mode": "sync"}
            if classifier_name in os.listdir(self.model_data_path):
                classifier_path = os.path.join(self.model_data_path, classifier_name)
                tile_classifier_model = model_api_Model.create_model(
                    model=classifier_path + ".xml",
                    weights_path=classifier_path + ".bin",
                    core=core,
                    preload=True,
                )
                tiler_arguments.update({"tile_classifier_model": tile_classifier_model})
            self._tiler = tiler_type(**tiler_arguments)
            self._tiling_enabled = True

        # TODO: This is a workaround to fix the issue that causes the output blob name
        #  to be unset. Remove this once it has been fixed on ModelAPI side
        output_names = list(self._inference_model.outputs.keys())
        if hasattr(self._inference_model, "output_blob_name"):
            if not self._inference_model.output_blob_name:
                self._inference_model.output_blob_name = {
                    name: name for name in output_names
                }

        # Force reload model to account for any postprocessing changes that may have
        # been applied while creating the ModelAPI wrapper
        logging.info(
            f"Inference model wrapper initialized, force reloading model on device "
            f"`{device}` to finalize inference model initialization process."
        )
        self._inference_model.load(force=True)

    @classmethod
    def from_model_and_hypers(
        cls, model: OptimizedModel, hyper_parameters: Optional[TaskConfiguration] = None
    ) -> "DeployedModel":
        """
        Create a DeployedModel instance out of an OptimizedModel and it's
        corresponding set of hyper parameters.

        :param model: OptimizedModel to convert to a DeployedModel
        :param hyper_parameters: TaskConfiguration instance containing the hyper
            parameters for the model
        :return: DeployedModel instance
        """
        model_dict = model.to_dict()
        model_dict.update({"hyper_parameters": hyper_parameters})
        model_group_id = model_dict.pop("model_group_id", None)
        deployed_model = cls(**model_dict)
        deployed_model.model_group_id = model_group_id
        try:
            base_url = model.base_url
            deployed_model.base_url = base_url
        except ValueError:
            pass
        return deployed_model

    @classmethod
    def from_folder(cls, path_to_folder: Union[str, os.PathLike]) -> "DeployedModel":
        """
        Create a DeployedModel instance from a folder containing the model data.

        :param path_to_folder: Path to the folder that holds the model data
        :return: DeployedModel instance
        """
        config_filepath = os.path.join(path_to_folder, "hyper_parameters.json")
        if os.path.isfile(config_filepath):
            with open(config_filepath, "r") as config_file:
                config_dict = json.load(config_file)
            hparams = ConfigurationRESTConverter.task_configuration_from_dict(
                config_dict
            )
        else:
            hparams = None
        model_detail_path = os.path.join(path_to_folder, "model.json")
        with open(model_detail_path, "r") as model_detail_file:
            model_detail_dict = json.load(model_detail_file)
        model = ModelRESTConverter.optimized_model_from_dict(model_detail_dict)
        deployed_model = cls.from_model_and_hypers(
            model=model, hyper_parameters=hparams
        )
        deployed_model.get_data(source=path_to_folder)
        return deployed_model

    def save(self, path_to_folder: Union[str, os.PathLike]) -> bool:
        """
        Save the DeployedModel instance to the designated folder.

        :param path_to_folder: Path to the folder to save the model to
        :return: True if the model was saved successfully, False otherwise
        """
        if self._model_data_path is None:
            raise ValueError(
                "No model definition and model weights data was found for {self}, "
                "unable to save model."
            )
        os.makedirs(path_to_folder, exist_ok=True, mode=0o770)

        new_model_data_path = os.path.join(path_to_folder, MODEL_DIR_NAME)
        new_model_python_path = os.path.join(path_to_folder, PYTHON_DIR_NAME)

        if (
            new_model_python_path == self._model_python_path
            or new_model_data_path == self._model_data_path
        ):
            logging.warning(
                f"Model '{self.name}' already exist in target path {path_to_folder}, "
                f"please save to a different location."
            )
            return False

        shutil.copytree(
            src=self._model_data_path,
            dst=new_model_data_path,
            dirs_exist_ok=True,
        )
        if self._model_python_path is not None:
            shutil.copytree(
                src=self._model_python_path,
                dst=new_model_python_path,
                dirs_exist_ok=True,
            )
            self._model_python_path = new_model_python_path

        self._model_data_path = new_model_data_path

        config_dict = ConfigurationRESTConverter.configuration_to_minimal_dict(
            self.hyper_parameters
        )
        config_filepath = os.path.join(path_to_folder, "hyper_parameters.json")
        with open(config_filepath, "w") as config_file:
            json.dump(config_dict, config_file, indent=4)

        model_detail_dict = self.to_dict()
        model_detail_dict.pop("hyper_parameters")
        model_detail_path = os.path.join(path_to_folder, "model.json")
        with open(model_detail_path, "w") as model_detail_file:
            json.dump(model_detail_dict, model_detail_file, indent=4)
        return True

    def _preprocess(
        self, image: np.ndarray
    ) -> Tuple[Dict[str, np.ndarray], Dict[str, Tuple[int, int, int]]]:
        """
        Preprocess an image for inference.

        :param image: Numpy array containing pixel data. The image is expected to have
            dimensions [height x width x channels], with the channels in RGB order
        :return: Dictionary holding the preprocessing result, the original shape of
            the image and the shape after preprocessing
        """
        return self._inference_model.preprocess(image)

    def _postprocess(
        self,
        inference_results: Dict[str, np.ndarray],
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Union[np.ndarray, List[Tuple[int, float]], Any]:
        """
        Postprocess the model outputs.

        :param inference_results: Dictionary holding the results of inference
        :param metadata: Dictionary holding metadata
        :return: Postprocessed inference results. The exact format depends on the
            type of model that is loaded:

            * For segmentation models, it will be a numpy array holding the output mask
            * For classification models, it will be a list of tuples holding the
                output class index and class probability
            * For detection models, it will be an instance of
                `openvino.model_zoo.model_api.models.utils.Detection`, holding the
                bounding box output
        """
        return self._inference_model.postprocess(inference_results, metadata)

    def infer(self, image: np.ndarray, explain: bool = False) -> Prediction:
        """
        Run inference on an already preprocessed image.

        :param image: numpy array representing an image
        :param explain: True to include saliency maps and feature maps in the returned
            Prediction. Note that these are only available if supported by the model.
        :return: Dictionary containing the model outputs
        """
        if not self._tiling_enabled:
            preprocessed_image, metadata = self._preprocess(image)
            # metadata is a dict with keys 'original_shape' and 'resized_shape'
            inference_results: Dict[str, np.ndarray] = self._inference_model.infer_sync(
                preprocessed_image
            )
        else:
            inference_results = self._tiler(image)
            metadata = {"original_shape": image.shape}
        return self._apply_postprocessing_steps(
            inference_results=inference_results, metadata=metadata, explain=explain
        )

    def infer_async(
        self,
        image: np.ndarray,
        explain: bool = False,
        runtime_data: Optional[Any] = None,
    ) -> None:
        """
        Perform asynchronous inference on the `image`.

        **NOTE**: Inference results are not returned directly! Instead, a
        post-inference callback should be defined to handle results, using the
        `.set_asynchronous_callback` method.

        :param image: numpy array representing an image
        :param explain: True to include saliency maps and feature maps in the returned
            Prediction. Note that these are only available if supported by the model.
        :param runtime_data: An optional object containing any additional data.
            that should be passed to the asynchronous callback for each infer request.
            This can for example be a timestamp or filename for the image to infer.
            You can for instance pass a dictionary, or a tuple/list of objects.
        """
        if not self._async_callback_defined:
            logging.warning(
                "No callback function defined to handle asynchronous inference, "
                "please make sure to define a callback using "
                "`.set_asynchronous_callback`, otherwise your inference results may be "
                "lost."
            )
        preprocessed_image, metadata = self._preprocess(image)
        self._inference_model.infer_async_raw(
            preprocessed_image,
            {"explain": explain, "metadata": metadata, "runtime_data": runtime_data},
        )

    def set_asynchronous_callback(
        self, callback_function: Callable[[Prediction, Optional[Any]], None]
    ) -> None:
        """
        Set the callback function to handle asynchronous inference results. This
        function is called whenever a result for an asynchronous inference request
        comes available.

        :param callback_function: Function that should be called to handle
            asynchronous inference results. The function should take the following
            input parameters:

             1. The inference results (the Prediction). This is the primary input
             2. Any additional data that will be passed with the infer
                request at runtime. For example, this could be a timestamp for the
                frame, or a title/filepath, etc. This can be in the form of any object:
                You can for instance pass a dictionary, or a tuple/list of multiple
                objects

        """
        if self._tiling_enabled:
            raise ValueError(
                "Asynchronous inference mode is not supported with models that use "
                "Tiling. Please use the synchronous inference mode instead."
            )

        def full_callback(infer_request, async_metadata: Dict[str, Any]):
            # Basic postprocessing, convert to `Prediction` object
            metadata = async_metadata["metadata"]
            explain = async_metadata["explain"]
            runtime_data = async_metadata["runtime_data"]

            raw_result = self._inference_model.inference_adapter.get_raw_result(
                infer_request
            )
            prediction = self._apply_postprocessing_steps(raw_result, metadata, explain)
            # User defined callback to further process the prediction results
            callback_function(prediction, runtime_data)

        self._inference_model.inference_adapter.set_callback(full_callback)
        self._async_callback_defined = True

    @property
    def label_schema(self) -> LabelSchema:
        """
        Return the LabelSchema for the model.

        This requires the inference model to be loaded, getting this property while
        inference models are not loaded will raise a ValueError

        :return: LabelSchema containing the SDK label schema for the model
        """
        if self._label_schema is None:
            raise ValueError(
                "Inference model is not loaded, unable to retrieve label schema. "
                "Please load inference model first."
            )
        return self._label_schema

    def _parse_label_schema_from_dict(
        self, label_schema_dict: Optional[Dict[str, Union[dict, List[dict]]]] = None
    ) -> None:
        """
        Parse the dictionary contained in the model `config.json` file, and
        generate an  LabelSchema from it.

        :param label_schema_dict: Dictionary containing the label schema information
            to parse
        """
        label_groups_list = label_schema_dict[LABEL_GROUPS_KEY]
        labels_dict = label_schema_dict[ALL_LABELS_KEY]
        for key, value in labels_dict.items():
            color_tuple = tuple(
                int(value["color"][key]) for key in ["red", "green", "blue"]
            )
            label_entity = Label(
                name=value["name"],
                color=rgb_to_hex(color_tuple),
                group=None,
                is_empty=value.get("is_empty", False),
                hotkey=value["hotkey"],
                domain=Domain[value["domain"]],
                id=value["_id"],
                is_anomalous=value.get("is_anomalous", False),
            )
            labels_dict[key] = label_entity
        label_groups: List[LabelGroup] = []
        for group_dict in label_groups_list:
            labels: List[Label] = [
                labels_dict[label_id] for label_id in group_dict["label_ids"]
            ]
            label_groups.append(
                LabelGroup(
                    id=group_dict["_id"],
                    name=group_dict["name"],
                    group_type=LabelGroupType[group_dict["relation_type"]],
                    labels=labels,
                )
            )
        self._label_schema = LabelSchema(label_groups=label_groups)

    def _apply_postprocessing_steps(
        self, inference_results: Any, metadata: Dict[str, Any], explain: bool
    ) -> Prediction:
        """
        Apply the required postprocessing steps to convert the model output to a
        Prediction object, with saliency maps and feature vector included if needed.

        :param inference_results: The results of the model
        :param metadata: Dictionary containing metadata about the original image
        :param explain: True to enable XAI outputs (saliency map and feature vector)
        :return: Prediction object containing the model predictions
        """
        if not self._tiling_enabled:
            postprocessing_results = self._postprocess(
                inference_results, metadata=metadata
            )
        else:
            postprocessing_results = inference_results

        prediction = self._converter.convert_to_prediction(
            postprocessing_results, image_shape=metadata["original_shape"]
        )

        # Add optional explainability outputs
        if explain:
            if self.has_xai_head:
                if hasattr(postprocessing_results, "feature_vector"):
                    prediction.feature_vector = postprocessing_results.feature_vector
                result_medium = ResultMedium(name="saliency map", type="saliency map")
                result_medium.data = self._converter.convert_saliency_map(
                    postprocessing_results, image_shape=metadata["original_shape"]
                )
                prediction.maps.append(result_medium)
            else:
                raise ValueError(
                    "Explainability outputs are not available for this model. "
                    "Please ensure the model has an explainability head."
                )
        return prediction

    def infer_queue_full(self) -> bool:
        """
        Return True if the queue for asynchronous infer requests is full, False
        otherwise

        :return: True if the infer queue is full, False otherwise
        """
        return not self._inference_model.inference_adapter.is_ready()

    def await_all(self) -> None:
        """
        Block execution untill all asynchronous infer requests have finished
        processing.

        This means that program execution will resume once the infer queue is empty

        This is a flow control function, it is only applicable when using
        asynchronous inference.
        """
        self._inference_model.inference_adapter.await_all()

    def await_any(self) -> None:
        """
        Block execution untill any of the asynchronous infer requests currently in
        the infer queue completes processing

        This means that program execution will resume once a single spot becomes
        available in the infer queue

        This is a flow control function, it is only applicable when using
        asynchronous inference.
        """
        self._inference_model.inference_adapter.await_any()

    @property
    def asynchronous_mode(self):
        """
        Return True if the DeployedModel is in asynchronous inference mode, False
        otherwise
        """
        return self._async_callback_defined

    @asynchronous_mode.setter
    def asynchronous_mode(self, mode: bool):
        """
        Set the DeployedModel to synchronous or asynchronous inference mode
        """
        if mode:
            if not self._async_callback_defined:
                raise ValueError(
                    "Please use the method `.set_asynchronous_callback()` to define a "
                    "callback and set the DeployedModel to asynchronous inference "
                    "mode."
                )
            else:
                logging.debug("DeployedModel is already in asynchronous mode")
        else:

            def do_nothing(request, userdata):
                pass

            self._async_callback_defined = False
            self._inference_model.inference_adapter.set_callback(do_nothing)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
from typing import Any, Dict, List

from geti_sdk.data_models import Annotation, Prediction
from geti_sdk.data_models.predictions import ResultMedium
from geti_sdk.rest_converters.annotation_rest_converter import (
    NormalizedAnnotationRESTConverter,
)

from .prediction_rest_converter import PredictionRESTConverter


class NormalizedPredictionRESTConverter(PredictionRESTConverter):
    """
    Class containing methods for converting predictions in normalized format to
    and from their REST representation

    It is a legacy class to support the annotation format in a normalized coordinate
    system, which was used in SCv1.1 and below
    """

    @staticmethod
    def normalized_prediction_from_dict(
        prediction: Dict[str, Any], image_width: int, image_height: int
    ) -> Prediction:
        """
        Legacy method that creates an AnnotationScene object from a dictionary
        returned by the /annotations REST endpoint in Intel® Geti™ versions 1.1 or
        below

        :param prediction: dictionary representing a Prediction, which
            contains all predictions for a certain media entity
        :param image_width: Width of the image to which the annotation scene applies
        :param image_height: Height of the image to which the annotation scene applies
        :return: Prediction object
        """
        input_copy = copy.deepcopy(prediction)
        media_identifier = (
            NormalizedAnnotationRESTConverter._media_identifier_from_dict(
                prediction["media_identifier"]
            )
        )
        annotations: List[Annotation] = []
        for annotation in prediction["annotations"]:
            annotations.append(
                NormalizedAnnotationRESTConverter.normalized_annotation_from_dict(
                    input_dict=annotation,
                    image_width=image_width,
                    image_height=image_height,
                )
            )
        result_media: List[ResultMedium] = []
        for result_medium in prediction.get("maps", []):
            if not isinstance(result_medium, ResultMedium):
                result_media.append(ResultMedium(**result_medium))
            else:
                result_media.append(result_medium)
        input_copy.update(
            {
                "annotations": annotations,
                "media_identifier": media_identifier,
                "maps": result_media,
            }
        )
        return Prediction(**input_copy)

    @staticmethod
    def to_normalized_dict(
        prediction: Prediction,
        image_width: int,
        image_height: int,
        deidentify: bool = True,
    ) -> Dict[str, Any]:
        """
        Convert a Prediction to a dictionary. By default, removes any ID
        fields in the output dictionary

        :param prediction: Prediction object to convert
        :param image_width:
        :param image_height:
        :param deidentify: True to remove any unique database ID fields in the output,
            False to keep these fields. Defaults to True
        :return: Dictionary holding the serialized AnnotationScene data
        """
        return NormalizedAnnotationRESTConverter.to_normalized_dict(
            annotation_scene=prediction,
            image_height=image_height,
            image_width=image_width,
            deidentify=deidentify,
        )



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
from typing import List, Optional

import numpy as np

from geti_sdk.data_models import Prediction
from geti_sdk.deployment.inference_hook_interfaces import PostInferenceTrigger


class ConfidenceTrigger(PostInferenceTrigger):
    """
    Post inference trigger based on the confidence of the predictions for an image.
    If the prediction confidence is **lower** than the defined `threshold`, the trigger
    is activated.

    Optionally a list of `label_names` can be defined, if passed then only the
    confidence level for the specified labels will be considered

    :param threshold: Confidence threshold to consider. Any predictions with a
        confidence **lower** than this threshold will activate the trigger
    :param label_names: Optional list of label names to include. If passed, only the
        confidences for the specified labels will be considered.
    """

    def __init__(self, threshold: float = 0.5, label_names: Optional[List[str]] = None):
        super().__init__(threshold=threshold)
        self.label_names = label_names

        if label_names is not None:
            self._repr_info_ += f", label_names={label_names}"

    def __call__(self, image: np.ndarray, prediction: Prediction) -> float:
        """
        Compute a trigger score for the `image` and corresponding `prediction`.

        :param image: Numpy array representing an image
        :param prediction: Prediction object corresponding to the inference result
            for the image.
        :return: Float representing the score for the input
        """
        min_confidence: float = 1
        for predicted_object in prediction.annotations:
            for label in predicted_object.labels:
                if self.label_names is not None:
                    if label.name not in self.label_names:
                        continue
                if label.probability < min_confidence:
                    min_confidence = label.probability
        return min_confidence

    def get_decision(self, score: float) -> bool:
        """
        Make a decision based on a previously computed `score` and the threshold defined
        for the trigger

        :param score: Float representing the score for a certain image, prediction pair
        :return: Boolean indicating whether the trigger conditions are met (True,
            score is **lower** than the threshold), or not (False)
        """
        return score < self.threshold



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
import json
import logging
from pprint import pformat
from typing import Any, ClassVar, Dict, List, Optional

import attr

from geti_sdk.data_models.enums import ModelStatus, OptimizationType
from geti_sdk.data_models.utils import (
    attr_value_serializer,
    deidentify,
    remove_null_fields,
    str_to_datetime,
    str_to_enum_converter,
)
from geti_sdk.utils.serialization_helpers import deserialize_dictionary

from .label import Label
from .performance import Performance


@attr.define
class OptimizationCapabilities:
    """
    Representation of the various model optimization capabilities in GETi.
    """

    is_nncf_supported: bool
    is_filter_pruning_supported: Optional[bool] = None
    is_filter_pruning_enabled: Optional[bool] = None


@attr.define
class OptimizationConfigurationParameter:
    """
    Representation of a parameter for model optimization in Geti.
    """

    name: str
    value: Any


@attr.define(slots=False)
class BaseModel:
    """
    Representation of the basic information for a Model or OptimizedModel in GETi
    """

    _identifier_fields: ClassVar[str] = [
        "id",
        "previous_revision_id",
        "previous_trained_revision_id",
    ]

    name: str
    fps_throughput: str
    latency: str
    precision: List[str]
    creation_date: str = attr.field(converter=str_to_datetime)
    size: Optional[int] = None
    target_device: Optional[str] = None
    target_device_type: Optional[str] = None
    previous_revision_id: Optional[str] = None
    previous_trained_revision_id: Optional[str] = None
    score: Optional[float] = attr.field(default=None)  # 'score' is removed in v1.1
    performance: Optional[Performance] = None
    id: Optional[str] = attr.field(default=None)
    label_schema_in_sync: Optional[bool] = attr.field(default=None)  # Added in Geti 1.1

    def __attrs_post_init__(self):
        """
        Initialize private attributes.
        """
        self._model_group_id: Optional[str] = None
        self._base_url: Optional[str] = None

    @property
    def model_group_id(self) -> Optional[str]:
        """
        Return the unique database ID of the model group to which the model belongs,
        if available.

        :return: ID of the model group for the model
        """
        return self._model_group_id

    @model_group_id.setter
    def model_group_id(self, id_: str):
        """
        Set the model group id for this model.

        :param id_: ID to set
        """
        self._model_group_id = id_

    @property
    def base_url(self) -> Optional[str]:
        """
        Return the base url that can be used to get the model details, download the
        model, etc., if available.

        :return: base url at which the model can be addressed. The url is defined
            relative to the ip address or hostname of the Intel® Geti™ server
        """
        if self._base_url is not None:
            return self._base_url
        else:
            raise ValueError(
                f"Insufficient data to determine base url for model {self}. Please "
                f"make sure that property `base_url` is set first."
            )

    @base_url.setter
    def base_url(self, base_url: str):
        """
        Set the base url that can be used to get the model details, download the
        model, etc.

        :param base_url: base url at which the model can be addressed
        :return:
        """
        if self.model_group_id is not None:
            if self.model_group_id in base_url:
                if base_url.endswith(f"models/{self.id}"):
                    base_url = base_url
                else:
                    base_url += f"/models/{self.id}"
            else:
                base_url += f"/{self.model_group_id}/models/{self.id}"
        else:
            base_url = base_url
        if hasattr(self, "optimized_models"):
            for model in self.optimized_models:
                model._base_url = base_url + f"/optimized_models/{model.id}"
        self._base_url = base_url

    def to_dict(self) -> Dict[str, Any]:
        """
        Return the dictionary representation of the model.

        :return:
        """
        base_dict = attr.asdict(
            self, recurse=True, value_serializer=attr_value_serializer
        )
        base_dict["model_group_id"] = self.model_group_id
        return base_dict

    @property
    def overview(self) -> str:
        """
        Return a string that represents an overview of the model.

        :return:
        """
        deidentified = copy.deepcopy(self)
        deidentified.deidentify()
        overview_dict = deidentified.to_dict()
        remove_null_fields(overview_dict)
        return pformat(overview_dict)

    def deidentify(self) -> None:
        """
        Remove unique database IDs from the BaseModel.
        """
        deidentify(self)


@attr.define(slots=False)
class OptimizedModel(BaseModel):
    """
    Representation of an OptimizedModel in Intel® Geti™. An optimized model is a trained model
    that has been converted OpenVINO representation. This conversion may involve weight
    quantization, filter pruning, or other optimization techniques supported by
    OpenVINO.
    """

    model_status: str = attr.field(
        kw_only=True, converter=str_to_enum_converter(ModelStatus)
    )
    optimization_methods: List[str] = attr.field(kw_only=True)
    optimization_objectives: Dict[str, Any] = attr.field(kw_only=True)
    optimization_type: str = attr.field(
        kw_only=True, converter=str_to_enum_converter(OptimizationType)
    )
    version: Optional[int] = attr.field(kw_only=True, default=None)
    configurations: Optional[List[OptimizationConfigurationParameter]] = attr.field(
        kw_only=True, default=None
    )  # Added in Geti v1.4
    model_format: Optional[str] = None  # Added in Geti v1.5
    has_xai_head: bool = False  # Added in Geti v1.5


@attr.define(slots=False)
class Model(BaseModel):
    """
    Representation of a trained Model in Intel® Geti™.
    """

    architecture: str = attr.field(kw_only=True)
    score_up_to_date: bool = attr.field(kw_only=True)
    optimization_capabilities: OptimizationCapabilities = attr.field(kw_only=True)
    optimized_models: List[OptimizedModel] = attr.field(kw_only=True)
    labels: Optional[List[Label]] = None
    version: Optional[int] = attr.field(default=None, kw_only=True)
    # 'version' is deprecated in v1.1 -- IS IT?
    training_dataset_info: Optional[Dict[str, str]] = None

    @property
    def model_group_id(self) -> Optional[str]:
        """
        Return the unique database ID of the model group to which the model belongs,
        if available.

        :return: ID of the model group for the model
        """
        return self._model_group_id

    @model_group_id.setter
    def model_group_id(self, id_: str):
        """
        Set the model group id for this model.

        :param id: ID to set
        """
        self._model_group_id = id_
        for model in self.optimized_models:
            model.model_group_id = id_

    @classmethod
    def from_dict(cls, model_dict: Dict[str, Any]) -> "Model":
        """
        Create a Model instance from a dictionary holding the model data.

        :param model_dict: Dictionary representing a model
        :return: Model instance reflecting the data contained in `model_dict`
        """
        return deserialize_dictionary(model_dict, cls)

    @classmethod
    def from_file(cls, filepath: str) -> "Model":
        """
        Create a Model instance from a .json file holding the model data.

        :param filepath: Path to a json file holding the model data
        :return:
        """
        with open(filepath, "r") as file:
            model_dict = json.load(file)
        return cls.from_dict(model_dict=model_dict)

    def get_optimized_model(
        self,
        optimization_type: Optional[str] = None,
        precision: Optional[str] = None,
        require_xai: bool = False,
    ) -> Optional[OptimizedModel]:
        """
        Return the OptimizedModel of the specified `optimization_type` or `precision`.
        The following optimization types are supported: 'nncf', 'pot', 'onnx', 'mo',
        `openvino`
        (case insensitive). The supported precision levels are `FP32`, `FP16`, `INT8`.
        Precision, optimization type or both can be specified.

        :param optimization_type: Optimization type for which to return the model
        :param precision: Precision level for which to return the model. Can be `FP32`,
            `FP16` or `INT8`
        :param require_xai: If True, only include models that have an XAI head for
            saliency map generation. Defaults to False
        :return: OptimizedModel object representing the optimized model
        """
        if optimization_type is None and precision is None:
            raise ValueError("Please specify optimization_type or precision, or both")
        optimized_models: List[OptimizedModel] = []
        if optimization_type is not None:
            capitalized_ot = optimization_type.upper()
            if capitalized_ot == "OPENVINO":
                optimized_models = [
                    model for model in self.optimized_models if "ONNX" not in model.name
                ]
            else:
                allowed_types = [item.name for item in OptimizationType]
                if capitalized_ot not in allowed_types:
                    raise ValueError(
                        f"Invalid optimization type passed, supported values are "
                        f"{allowed_types} or `openvino`"
                    )
                optimization_type = OptimizationType(capitalized_ot)
                optimized_models = [
                    model
                    for model in self.optimized_models
                    if model.optimization_type == optimization_type
                ]

        if precision is not None:
            if len(optimized_models) == 0:
                models_to_search = self.optimized_models
            else:
                models_to_search = optimized_models
            optimized_models = [
                model for model in models_to_search if precision in model.name
            ]

        if len(optimized_models) == 0:
            logging.info(
                "No optimized model meeting the optimization criteria was found."
            )
            return None
        elif len(optimized_models) == 1:
            if not require_xai:
                return optimized_models[0]
            else:
                if optimized_models[0].has_xai_head:
                    return optimized_models[0]
                logging.info(
                    f"An optimized model of type {optimization_type} was found, but it does "
                    f"not include an XAI head. Method `get_optimized_model` returned "
                    f"None."
                )
                return None
        else:
            if not require_xai:
                models_to_check = optimized_models
            else:
                models_to_check = [m for m in optimized_models if m.has_xai_head]
            if len(models_to_check) == 0:
                logging.info(
                    f"An optimized model of type {optimization_type} was found, but it does "
                    f"not include an XAI head. Method `get_optimized_model` returned "
                    f"None."
                )
                return None
            creation_dates = [om.creation_date for om in models_to_check]
            max_index = creation_dates.index(max(creation_dates))
            return models_to_check[max_index]



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import copy
import logging
import warnings
from datetime import datetime
from typing import Any, Dict, Optional, Union

import attrs
import cv2
import numpy as np

from geti_sdk.data_models import Dataset, Prediction, Project
from geti_sdk.deployment.inference_hook_interfaces import PostInferenceAction
from geti_sdk.http_session import (
    GetiRequestException,
    GetiSession,
    ServerCredentialConfig,
    ServerTokenConfig,
)
from geti_sdk.rest_clients.dataset_client import DatasetClient
from geti_sdk.rest_clients.media_client.image_client import ImageClient
from geti_sdk.rest_clients.project_client.project_client import ProjectClient


class GetiDataCollection(PostInferenceAction):
    """
    Post inference action that will send an image to a specified `project` and `dataset`
    on the Intel® Geti™ server addressed by `session`.

    :param session: Geti session representing the connecting to the Intel® Geti™ server
    :param workspace_id: unique ID of the workspace in which the project to collect
        the data resides.
    :param project: Project or name of the project to whicht the image data should be
        transferred.
    :param dataset: Optional Dataset or name of the dataset in which to store the
        image data. If not specified, the default training dataset of the project is
        used
    :param log_level: Log level for the action. Options are 'info' or 'debug'
    """

    _override_from_dict_: bool = True

    def __init__(
        self,
        session: GetiSession,
        workspace_id: str,
        project: Union[str, Project],
        dataset: Optional[Union[str, Dataset]] = None,
        log_level: str = "debug",
    ):
        super().__init__(log_level=log_level)
        project_client = ProjectClient(session=session, workspace_id=workspace_id)
        if isinstance(project, str):
            project_name = project
            project = project_client.get_project_by_name(project_name=project_name)
            if project is None:
                raise ValueError(
                    f"Project `{project_name}` does not exist on the Intel® Geti™ "
                    f"server, unable to initialize the Intel® Geti™ data collection "
                    f"action"
                )
        dataset_client = DatasetClient(
            session=session, workspace_id=workspace_id, project=project
        )
        if dataset is None:
            datasets = dataset_client.get_all_datasets()
            dataset = [ds for ds in datasets if ds.use_for_training][0]
        elif isinstance(dataset, str):
            dataset_name = dataset

            try:
                dataset = dataset_client.get_dataset_by_name(dataset_name)
            except ValueError:
                dataset = dataset_client.create_dataset(dataset_name)
                self.log_function(
                    f"Dataset `{dataset_name}` was created in project `{project.name}`"
                )
        self.image_client = ImageClient(
            session=session, workspace_id=workspace_id, project=project
        )
        self.dataset = dataset
        self._repr_info_ = (
            f"target_server=`{session.config.host}`, "
            f"target_project={project.name}, "
            f"target_dataset={dataset.name}"
        )

        # Serialize input arguments
        self._constructor_arguments_["project"] = project.name
        self._constructor_arguments_["dataset"] = dataset.name
        self._constructor_arguments_["session"] = attrs.asdict(session.config)

    def __call__(
        self,
        image: np.ndarray,
        prediction: Prediction,
        score: Optional[float] = None,
        name: Optional[str] = None,
        timestamp: Optional[datetime] = None,
    ):
        """
        Execute the action, upload the given `image` to the Intel® Geti™ server.

        The parameters `prediction`, `score`, `name` and `timestamp` are not used in
        this specific action.

        :param image: Numpy array representing an image
        :param prediction: Prediction object which was generated for the image
        :param score: Optional score computed from a post inference trigger
        :param name: String containing the name of the image
        :param timestamp: Datetime object containing the timestamp belonging to the
            image
        """
        image_bgr = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        # upload_image uses cv2 to encode the numpy array as image, so it expects an
        # image in BGR format. However, `Deployment.infer` requires RGB format, so
        # we have to convert
        try:
            self.image_client.upload_image(image=image_bgr, dataset=self.dataset)
        except GetiRequestException as e:
            logging.exception(e)
        self.log_function(
            f"GetiDataCollection inference action uploaded image to dataset "
            f"`{self.dataset.name}`"
        )

    @classmethod
    def from_dict(cls, input_dict: Dict[str, Any]) -> "GetiDataCollection":
        """
        Construct a GetiDataCollection post inference action object from an input
        dictionary `input_dict`

        :param input_dict: Dictionary representation of the GetiDataCollection
        :return: Instantiated GetiDataCollection, according to the input dictionary
        """
        input_copy = copy.deepcopy(input_dict)
        session_dict = input_copy["session"]
        if "username" in session_dict:
            server_config_class = ServerCredentialConfig
        elif "token" in session_dict:
            server_config_class = ServerTokenConfig
        else:
            raise ValueError(
                f"Invalid `GetiSession` parameters encountered: {session_dict}"
            )
        session = GetiSession(server_config_class(**session_dict))
        input_copy.update({"session": session})
        return cls(**input_copy)

    def to_dict(self) -> Dict[str, Any]:
        """
        Return a dictionary representation of the PostInferenceObject

        :return: Dictionary representing the class name and its constructor parameters
        """
        warnings.warn(
            "GetiDataCollection post inference action contains sensitive information "
            "used for authentication on the Intel® Geti™ platform. Be careful when "
            "saving this information to disk or sharing with others!"
        )
        return super().to_dict()



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
from typing import List

from pathvalidate import sanitize_filename

from geti_sdk.data_models.enums import TaskType
from geti_sdk.data_models.project import Project


def get_task_types_by_project_type(project_type: str) -> List[TaskType]:
    """
    Return a list of task_type for each task in the project pipeline, for a certain
    'project_type'.

    :param project_type:
    :return:
    """
    return [TaskType(task) for task in project_type.split("_to_")]


def get_project_folder_name(project: Project) -> str:
    """
    Return a folder name for the project, that can be used to download the project to.

    :param project: Project to get a folder name for
    :return: string holding the folder name for the project
    """
    name_part = sanitize_filename(project.name)
    return f"{project.id}_{name_part}"



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
from typing import Generic, List, Optional, Sequence, Union

from geti_sdk.data_models import AnnotationScene, Image, Video, VideoFrame
from geti_sdk.data_models.containers import MediaList
from geti_sdk.http_session import GetiRequestException

from .base_annotation_client import AnnotationReaderType, BaseAnnotationClient


class AnnotationClient(BaseAnnotationClient, Generic[AnnotationReaderType]):
    """
    Class to up- or download annotations for images or videos to an existing project.
    """

    def get_latest_annotations_for_video(self, video: Video) -> List[AnnotationScene]:
        """
        Retrieve all latest annotations for a video from the cluster.

        If the video does not have any annotations yet, this method returns an
        empty list

        :param video: Video to get the annotations for
        :return: List of AnnotationScene's, each entry corresponds to an
            AnnotationScene for a single frame in the video
        """
        try:
            response = self.session.get_rest_response(
                url=f"{video.base_url}/annotations/latest",
                method="GET",
                include_organization_id=False,
            )
        except GetiRequestException as error:
            if error.status_code == 204:
                return []
            else:
                raise error
        if self.session.version.is_sc_1_1 or self.session.version.is_sc_mvp:
            annotations = response
        else:
            annotations = response["video_annotations"]
        annotation_scenes = [
            self.annotation_scene_from_rest_response(
                annotation_scene, media_information=video.media_information
            )
            for annotation_scene in annotations
            if annotation_scene["annotations"]
        ]
        for scene in annotation_scenes:
            scene.resolve_label_names_and_colors(labels=self._project.get_all_labels())
        return annotation_scenes

    def upload_annotations_for_video(
        self, video: Video, append_annotations: bool = False
    ):
        """
        Upload annotations for a video. If append_annotations is set to True,
        annotations will be appended to the existing annotations for the video in the
        project. If set to False, existing annotations will be overwritten.

        :param video: Video to upload annotations for
        :param append_annotations:
        :return:
        """
        annotation_filenames = self.annotation_reader.get_data_filenames()
        video_annotation_names = [
            filename
            for filename in annotation_filenames
            if filename.startswith(f"{video.name}_frame_")
        ]
        frame_indices = [int(name.split("_")[-1]) for name in video_annotation_names]
        video_frames = MediaList(
            [
                VideoFrame.from_video(video=video, frame_index=frame_index)
                for frame_index in frame_indices
            ]
        )
        upload_count = self._upload_annotations_for_2d_media_list(
            media_list=video_frames, append_annotations=append_annotations
        )
        return upload_count

    def upload_annotations_for_videos(
        self, videos: Sequence[Video], append_annotations: bool = False
    ):
        """
        Upload annotations for a list of videos. If append_annotations is set to True,
        annotations will be appended to the existing annotations for the video in the
        project. If set to False, existing annotations will be overwritten.

        :param videos: List of videos to upload annotations for
        :param append_annotations:
        :return:
        """
        logging.info("Starting video annotation upload...")
        upload_count = 0
        for video in videos:
            upload_count += self.upload_annotations_for_video(
                video=video, append_annotations=append_annotations
            )
        if upload_count > 0:
            logging.info(
                f"Upload complete. Uploaded {upload_count} new video frame annotations"
            )
        else:
            logging.info("No new video frame annotations were found.")

    def upload_annotations_for_images(
        self, images: Sequence[Image], append_annotations: bool = False
    ):
        """
        Upload annotations for a list of images. If append_annotations is set to True,
        annotations will be appended to the existing annotations for the image in the
        project. If set to False, existing annotations will be overwritten.

        :param images: List of images to upload annotations for
        :param append_annotations:
        :return:
        """
        logging.info("Starting image annotation upload...")
        upload_count = self._upload_annotations_for_2d_media_list(
            media_list=images, append_annotations=append_annotations
        )
        if upload_count > 0:
            logging.info(
                f"Upload complete. Uploaded {upload_count} new image annotations"
            )
        else:
            logging.info("No new image annotations were found.")

    def download_annotations_for_video(
        self,
        video: Video,
        path_to_folder: str,
        append_video_uid: bool = False,
        max_threads: int = 10,
    ) -> float:
        """
        Download video annotations from the server to a target folder on disk.

        :param video: Video for which to download the annotations
        :param path_to_folder: Folder to save the annotations to
        :param append_video_uid: True to append the UID of the video to the
             annotation filename (separated from the original filename by an underscore,
             i.e. '{filename}_{media_id}'). This can be useful if the project contains
             videos with duplicate filenames. If left as False, the video filename and
             frame index for the annotation are used as filename for the downloaded
             annotation.
        :param max_threads: Maximum number of threads to use for downloading. Defaults to 10.
            Set to -1 to use all available threads.
        :return: Returns the time elapsed to download the annotations, in seconds
        """
        annotations = self.get_latest_annotations_for_video(video=video)
        frame_list = MediaList[VideoFrame](
            [
                VideoFrame.from_video(
                    video=video, frame_index=annotation.media_identifier.frame_index
                )
                for annotation in annotations
            ]
        )
        if len(frame_list) > 0:
            return self._download_annotations_for_2d_media_list(
                media_list=frame_list,
                path_to_folder=path_to_folder,
                verbose=False,
                append_media_uid=append_video_uid,
                max_threads=max_threads,
            )
        else:
            return 0

    def download_annotations_for_images(
        self,
        images: MediaList[Image],
        path_to_folder: str,
        append_image_uid: bool = False,
        max_threads: int = 10,
    ) -> float:
        """
        Download image annotations from the server to a target folder on disk.

        :param images: List of images for which to download the annotations
        :param path_to_folder: Folder to save the annotations to
        :param append_image_uid: True to append the UID of the image to the
            annotation filename (separated from the original filename by an underscore,
             i.e. '{filename}_{media_id}'). This can be useful if the project contains
             images with duplicate filenames. If left as False, the image filename is
             used as filename for the downloaded annotation as well.
        :param max_threads: Maximum number of threads to use for downloading. Defaults to 10.
            Set to -1 to use all available threads.
        :return: Returns the time elapsed to download the annotations, in seconds
        """
        return self._download_annotations_for_2d_media_list(
            media_list=images,
            path_to_folder=path_to_folder,
            append_media_uid=append_image_uid,
            verbose=False,
            max_threads=max_threads,
        )

    def download_annotations_for_videos(
        self,
        videos: MediaList[Video],
        path_to_folder: str,
        append_video_uid: bool = False,
        max_threads: int = 10,
    ) -> float:
        """
        Download annotations for a list of videos from the server to a target folder
        on disk.

        :param videos: List of videos for which to download the annotations
        :param path_to_folder: Folder to save the annotations to
        :param append_video_uid: True to append the UID of the video to the
             annotation filename (separated from the original filename by an underscore,
             i.e. '{filename}_{media_id}'). This can be useful if the project contains
             videos with duplicate filenames. If left as False, the video filename and
             frame index for the annotation are used as filename for the downloaded
             annotation.
        :param max_threads: Maximum number of threads to use for downloading. Defaults to 10.
            Set to -1 to use all available threads.
        :return: Time elapsed to download the annotations, in seconds
        """
        t_total = 0
        logging.info(
            f"Starting annotation download... saving annotations for "
            f"{len(videos)} videos to folder {path_to_folder}/annotations"
        )
        for video in videos:
            t_total += self.download_annotations_for_video(
                video=video,
                path_to_folder=path_to_folder,
                append_video_uid=append_video_uid,
                max_threads=max_threads,
            )
        logging.info(f"Video annotation download finished in {t_total:.1f} seconds.")
        return t_total

    def download_all_annotations(
        self, path_to_folder: str, max_threads: int = 10
    ) -> None:
        """
        Download all annotations for the project to a target folder on disk.

        :param path_to_folder: Folder to save the annotations to
        :param max_threads: Maximum number of threads to use for downloading. Defaults to 10.
            Set to -1 to use all available threads.
        """
        image_list = self._get_all_media_by_type(media_type=Image)
        video_list = self._get_all_media_by_type(media_type=Video)
        if len(image_list) > 0:
            self.download_annotations_for_images(
                images=image_list,
                path_to_folder=path_to_folder,
                append_image_uid=image_list.has_duplicate_filenames,
                max_threads=max_threads,
            )
        if len(video_list) > 0:
            self.download_annotations_for_videos(
                video_list,
                path_to_folder=path_to_folder,
                append_video_uid=video_list.has_duplicate_filenames,
                max_threads=max_threads,
            )

    def upload_annotations_for_all_media(self, append_annotations: bool = False):
        """
        Upload annotations for all media in the project, If append_annotations is set
        to True, annotations will be appended to the existing annotations for the
        media on the server. If set to False, existing annotations will be overwritten.

        :param append_annotations: True to append annotations from the local disk to
            the existing annotations on the server, False to overwrite the server
            annotations by those on the local disk. Defaults to False.
        """
        image_list = self._get_all_media_by_type(media_type=Image)
        video_list = self._get_all_media_by_type(media_type=Video)
        if len(image_list) > 0:
            self.upload_annotations_for_images(
                images=image_list, append_annotations=append_annotations
            )
        if len(video_list) > 0:
            self.upload_annotations_for_videos(
                videos=video_list, append_annotations=append_annotations
            )

    def upload_annotation(
        self, media_item: Union[Image, VideoFrame], annotation_scene: AnnotationScene
    ) -> AnnotationScene:
        """
        Upload an annotation for an image or video frame to the Intel® Geti™ server.

        :param media_item: Image or VideoFrame to apply and upload the annotation to
        :param annotation_scene: AnnotationScene to upload
        :return: The uploaded annotation
        """
        if not isinstance(media_item, (Image, VideoFrame)):
            raise ValueError(
                f"Cannot upload annotation for media item {media_item.name}. This "
                f"method only supports uploading annotations for single images and "
                f"video frames. Please use the method `upload_annotations_for_video` "
                f"to upload video annotations"
            )
        return self._upload_annotation_for_2d_media_item(
            media_item=media_item, annotation_scene=annotation_scene
        )

    def get_annotation(
        self, media_item: Union[Image, VideoFrame]
    ) -> Optional[AnnotationScene]:
        """
        Retrieve the latest annotations for an image or video frame from the
        Intel® Geti™ platform.
        If no annotation is available, this method returns None.

        :param media_item: Image or VideoFrame to retrieve the annotations for
        :return: AnnotationScene instance containing the latest annotation data
        """
        if not isinstance(media_item, (Image, VideoFrame)):
            raise ValueError(
                f"Cannot get annotation for media item {media_item.name}. This method "
                f"only supports getting annotations for images and video frames. "
                f"Please use the method `get_latest_annotations_for_video` to retrieve "
                f"video annotations"
            )
        return self._get_latest_annotation_for_2d_media_item(media_item=media_item)



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import numpy as np

from geti_sdk.data_models import Prediction
from geti_sdk.deployment.inference_hook_interfaces import PostInferenceTrigger


class EmptyLabelTrigger(PostInferenceTrigger):
    """
    Post inference trigger that will activate if the prediction is empty
    """

    def __init__(self):
        # LabelTrigger will return a score of 1 if label is found, so we can use the
        # default threshold defined in the super class
        super().__init__()
        self._repr_info_ = ""

    def __call__(self, image: np.ndarray, prediction: Prediction) -> float:
        """
        Compute a trigger score for the `image` and corresponding `prediction`.

        :param image: Numpy array representing an image
        :param prediction: Prediction object corresponding to the inference result
            for the image.
        :return: Float representing the score for the input
        """
        for predicted_object in prediction.annotations:
            if len(predicted_object.labels) > 1:
                return 0
            if not predicted_object.shape.is_full_box(
                image_width=image.shape[1], image_height=image.shape[0]
            ):
                return 0
            for label in predicted_object.labels:
                if label.name.lower() == "no object":
                    return 1
        return 0



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Any, Dict

from geti_sdk.data_models import Job
from geti_sdk.utils import deserialize_dictionary


class JobRESTConverter:
    """
    Class that handles conversion of Intel® Geti™ REST output for jobs to objects,
    and vice-versa
    """

    @staticmethod
    def from_dict(job_dict: Dict[str, Any]) -> Job:
        """
        Create a Job instance from the input dictionary passed in `job_dict`.

        :param job_dict: Dictionary representing a job on the Intel® Geti™ server, as
            returned by the /jobs endpoints
        :return: Job instance, holding the job data contained in job_dict
        """
        # There is an inconsistency in the REST API, the `scores` field was changed
        # from array to object. Preprocess the data to account for that
        if "metadata" in job_dict.keys():
            metadata = job_dict["metadata"]
            if "scores" in metadata.keys():
                scores = metadata["scores"]
                if not isinstance(scores, list):
                    metadata["scores"] = [scores]

        return deserialize_dictionary(job_dict, output_type=Job)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# noqa: D104

from .anomaly_demos import (
    create_anomaly_classification_demo_project,
    ensure_trained_anomaly_project,
)
from .coco_demos import (
    create_classification_demo_project,
    create_detection_demo_project,
    create_detection_to_classification_demo_project,
    create_detection_to_segmentation_demo_project,
    create_segmentation_demo_project,
    ensure_trained_example_project,
)
from .utils import ensure_project_is_trained

__all__ = [
    "create_detection_demo_project",
    "create_classification_demo_project",
    "create_segmentation_demo_project",
    "create_detection_to_segmentation_demo_project",
    "create_detection_to_classification_demo_project",
    "create_anomaly_classification_demo_project",
    "ensure_trained_example_project",
    "ensure_trained_anomaly_project",
    "ensure_project_is_trained",
]



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from enum import Enum
from typing import List, Optional

from geti_sdk.data_models.label import Label


class LabelGroupType(Enum):
    """Enum to indicate the LabelGroupType."""

    EXCLUSIVE = 1
    EMPTY_LABEL = 2


class LabelGroup:
    """
    Representation of a group of labels.
    """

    def __init__(
        self,
        name: str,
        labels: List[Label],
        group_type: LabelGroupType = LabelGroupType.EXCLUSIVE,
        id: Optional[str] = None,
    ) -> None:
        """
        Initialize a LabelGroup object.

        :param name: The name of the label group.
        :param labels: A list of Label objects associated with the group.
        :param group_type: The type of the label group. Defaults to LabelGroupType.EXCLUSIVE.
        :param id: The ID of the label group. Defaults to None.
        """
        self.id = id
        self.name = name
        self.group_type = group_type
        self.labels = sorted(labels, key=lambda label: label.id)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from enum import Enum


class MediaType(Enum):
    """
    Enum representing the different media types in GETi.
    """

    IMAGE = "image"
    VIDEO = "video"
    VIDEO_FRAME = "video_frame"

    def __str__(self) -> str:
        """
        Return the string representation of the MediaType instance.
        """
        return self.value


SUPPORTED_VIDEO_FORMATS = [".mp4", ".avi", ".mkv", ".mov"]
SUPPORTED_IMAGE_FORMATS = [".jpg", ".jpeg", ".bmp", ".png", ".tif", ".tiff"]



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import json
import logging
import os
from typing import List, Optional, TypeVar, Union

from geti_sdk.data_models import (
    Algorithm,
    Job,
    Model,
    ModelGroup,
    ModelSummary,
    OptimizedModel,
    Project,
    Task,
)
from geti_sdk.data_models.enums import JobState, JobType, OptimizationType
from geti_sdk.http_session import GetiSession
from geti_sdk.rest_converters import ModelRESTConverter
from geti_sdk.utils import get_supported_algorithms
from geti_sdk.utils.job_helpers import get_job_with_timeout, monitor_job

ModelType = TypeVar("ModelType", Model, OptimizedModel)


class ModelClient:
    """
    Class to manage the models and model groups for a certain project
    """

    def __init__(self, workspace_id: str, project: Project, session: GetiSession):
        self.session = session
        project_id = project.id
        self.project = project
        self.workspace_id = workspace_id
        self.task_ids = [task.id for task in project.get_trainable_tasks()]
        self.base_url = f"workspaces/{workspace_id}/projects/{project_id}/model_groups"
        self.supported_algos = get_supported_algorithms(
            rest_session=self.session, project=project, workspace_id=workspace_id
        )

    def get_all_model_groups(self) -> List[ModelGroup]:
        """
        Return a list of all model groups in the project.

        :return: List of model groups in the project
        """
        response = self.session.get_rest_response(url=self.base_url, method="GET")
        if self.session.version.is_sc_1_1 or self.session.version.is_sc_mvp:
            # The API is not fully consistent here, depending on exact release.
            # Response may either be a dict or a list
            try:
                response_array = response["items"]
            except TypeError:
                response_array = response
        else:
            response_array = response["model_groups"]
        model_groups = [
            ModelRESTConverter.model_group_from_dict(group) for group in response_array
        ]
        # Update algorithm details
        for group in model_groups:
            group.algorithm = self.supported_algos.get_by_model_template(
                model_template_id=group.model_template_id
            )
            for model in group.models:
                # set the model storage id, to link models to their parent group
                model.model_storage_id = group.id
        return model_groups

    def get_latest_model_for_all_model_groups(self) -> List[Model]:
        """
        Return the latest trained models for each model group in the project.

        :return: List of models, one for each trained algorithm in the project.
        """
        model_groups = self.get_all_model_groups()
        latest_models: List[Model] = []
        for model_group in model_groups:
            lm = model_group.get_latest_model()
            latest_models.append(
                self._get_model_detail(group_id=model_group.id, model_id=lm.id)
            )
        return latest_models

    def get_model_group_by_algo_name(self, algorithm_name: str) -> Optional[ModelGroup]:
        """
        Return the model group for the algorithm named `algorithm_name`, if any. If
        no model group for this algorithm is found in the project, this method returns
        None

        :param algorithm_name: Name of the algorithm
        :return: ModelGroup instance corresponding to this algorithm
        """
        model_groups = self.get_all_model_groups()
        return next(
            (group for group in model_groups if group.algorithm.name == algorithm_name),
            None,
        )

    def get_latest_model_by_algo_name(self, algorithm_name: str) -> Optional[Model]:
        """
        Return the latest model for a specific algorithm. If no model has been trained
        for the algorithm, this method returns None.

        :param algorithm_name: Name fo the algorithm for which to return the model
        :return: Model object respresenting the model.
        """
        model_group = self.get_model_group_by_algo_name(algorithm_name)
        if model_group:
            model_summary = model_group.get_latest_model()
            return self._get_model_detail(
                group_id=model_group.id, model_id=model_summary.id
            )
        else:
            return None

    def get_latest_optimized_model(
        self,
        algorithm_name: str,
        optimization_type: str = "MO",
        precision: str = "FP16",
        require_xai: bool = False,
    ) -> OptimizedModel:
        """
        Return the optimized model for the latest trained model for a specified
        algorithm. Additional parameters allow filtering on the optimization type
        (e.g. 'nncf', 'pot', 'mo', 'onnx'), precision ('int8', 'fp16', 'fp32') and
        whether or not the model includes an XAI head for saliency map generation.

        If no optimized model for the specified criteria can be found, this method
        raises an error

        :param algorithm_name: Name of the algorithm to retrieve the model for
        :param optimization_type: Optimization type to select. Options are 'mo',
            'nncf', 'pot', 'onnx'. Case insensitive. Defaults to 'MO'
        :param precision: Model precision to select. Options are 'INT8', 'FP16', 'FP32'.
            Defaults to 'FP16'
        :param require_xai: If True, only select models that include an XAI head.
            Defaults to False
        """
        base_model = self.get_latest_model_by_algo_name(algorithm_name=algorithm_name)
        if base_model is None:
            raise RuntimeError(
                f"No trained model was found for algorithm `{algorithm_name}`"
            )
        n_optimized_models = len(base_model.optimized_models)
        logging.info(
            f"{n_optimized_models} optimized models were found for algorithm "
            f"`{algorithm_name}`. Finding the most recent optimized model with "
            f"precision {precision} and optimization type {optimization_type}."
        )
        if require_xai:
            opt_models = [m for m in base_model.optimized_models if m.has_xai_head]
        else:
            opt_models = base_model.optimized_models
        if n_optimized_models != 0 and len(opt_models) == 0:
            raise RuntimeError(
                f"Algorithm {algorithm_name} has a trained base model and "
                f"{n_optimized_models} optimized models, but no optimized model with "
                f"XAI head was found."
            )
        cap_prec = precision.upper()
        supported_precisions = ["FP32", "FP16", "INT8"]
        if cap_prec not in supported_precisions:
            raise ValueError(
                f"Invalid target precision specified: {precision}. Supported options "
                f"are: {supported_precisions}"
            )
        opt_models_precision = [om for om in opt_models if cap_prec in om.name]
        cap_opt_type = optimization_type.upper()
        opt_type = OptimizationType(cap_opt_type)
        opt_models_prec_type = [
            om for om in opt_models_precision if om.optimization_type == opt_type
        ]
        if len(opt_models_prec_type) == 0:
            raise RuntimeError(
                f"Algorithm {algorithm_name} has a trained base model and "
                f"{n_optimized_models} optimized models, but no optimized model "
                f"matches the required optimization type and precision."
            )
        elif len(opt_models_prec_type) == 1:
            return opt_models_prec_type[0]
        else:
            logging.info(
                f"Found {len(opt_models_prec_type)} models that match the selection "
                f"criteria. Returning the most recently created matching model."
            )
            creation_dates = [om.creation_date for om in opt_models_prec_type]
            max_index = creation_dates.index(max(creation_dates))
            return opt_models_prec_type[max_index]

    def get_model_by_algorithm_task_and_version(
        self,
        algorithm: Algorithm,
        version: Optional[int] = None,
        task: Optional[Task] = None,
    ) -> Optional[Model]:
        """
        Retrieve a Model from the Intel® Geti™ server, corresponding to a specific
        algorithm and model version. If no version is passed, this method will
        retrieve the latest model for the algorithm.

        If no model for the algorithm is available in the project, this method returns
        None

        :param algorithm: Algorithm for which to get the model
        :param version: Version of the model to retrieve. If left as None, returns the
            latest version
        :param task: Task for which to get the model. If left as None, this method
            searches for models for `algorithm` in all tasks in the project
        :return: Model object corresponding to `algorithm` and `version`, for a
            specific `task`, if any. If no model is found by those parameters, this
            method returns None
        """
        if task is not None:
            if algorithm.task_type != task.type:
                raise ValueError(
                    f"Unable to retrieve model. The algorithm {algorithm} is not "
                    f"available for the task {task}"
                )
        model_groups = self.get_all_model_groups()
        model_group: Optional[ModelGroup] = None
        for group in model_groups:
            if group.algorithm == algorithm:
                if task is None:
                    model_group = group
                    break
                else:
                    if group.task_id == task.id:
                        model_group = group
                        break
        if model_group is None:
            return None
        if version is not None:
            try:
                model_summary = model_group.get_model_by_version(version=version)
            except ValueError:
                return None
        else:
            model_summary = model_group.get_latest_model()
        return self._get_model_detail(model_group.id, model_id=model_summary.id)

    def _get_model_detail(self, group_id: str, model_id: str) -> Model:
        """
        Return the Model object holding detailed information about the model with
        `model_id` living in the model group with `group_id`.

        :param group_id: Unique database ID of the model group to which the model to
            get belongs
        :param model_id: Unique database ID of the model to get
        :return: Model instance holding detailed information about the model
        """
        model_detail = self.session.get_rest_response(
            url=f"{self.base_url}/{group_id}/models/{model_id}", method="GET"
        )
        model = ModelRESTConverter.model_from_dict(model_detail)
        model.model_group_id = group_id
        model.base_url = self.base_url
        return model

    def update_model_detail(self, model: Union[Model, ModelSummary]) -> Model:
        """
        Update the model such that its details are up to date. This includes updating
        the list of available optimized models for the model.

        :param model: Model or ModelSummary object, representing the model to update
        :return: Model object containing the up to date details of the model
        """
        if isinstance(model, ModelSummary):
            group_id = model.model_storage_id
        elif isinstance(model, Model):
            group_id = model.model_group_id
        else:
            raise TypeError(
                f"Invalid type {type(model)}. Argument `model` must be either a "
                f"Model or ModelSummary object"
            )
        model_detail = self.session.get_rest_response(
            url=f"{self.base_url}/{group_id}/models/{model.id}", method="GET"
        )
        updated_model = ModelRESTConverter.model_from_dict(model_detail)
        updated_model.model_group_id = group_id
        updated_model.base_url = self.base_url
        return updated_model

    def set_active_model(
        self,
        model: Optional[Union[Model, ModelSummary]] = None,
        algorithm: Optional[Union[Algorithm, str]] = None,
    ) -> None:
        """
        Set the model as the active model.

        :param model: Model or ModelSummary object representing the model to set as active
        :param algorithm: Algorithm or algorithm name for which to set the model as active
        :raises ValueError: If neither `model` nor `algorithm` is specified,
            If the algorithm is not supported in the project,
            If unable to set the active model
        """
        # First we determine the algorithm name
        if model is not None:
            # Update the model details to make sure we have the latest information
            model = self.update_model_detail(model)
            algorithm_name = model.architecture
        elif algorithm is not None:
            if isinstance(algorithm, str):
                algorithm_name = algorithm
            elif isinstance(algorithm, Algorithm):
                algorithm_name = algorithm.name
            else:
                raise ValueError(
                    f"Invalid type {type(algorithm)}. Argument `algorithm` must be "
                    "either a string representing the algorith name or an Algorithm object"
                )
        else:
            raise ValueError(
                "Either `model` or `algorithm` must be specified to set the active model"
            )
        # Now we make sure that the algorithm is supported in the project
        algorithms_supported_in_the_project = {
            algorithm.name
            for task in self.project.get_trainable_tasks()
            for algorithm in self.supported_algos.get_by_task_type(task.type)
        }
        if algorithm_name not in algorithms_supported_in_the_project:
            raise ValueError(
                f"Algorithm `{algorithm_name}` is not supported in the project "
                f"{self.project.name}."
            )
        # We get a model group for the algorithm
        model_group = self.get_model_group_by_algo_name(algorithm_name=algorithm_name)
        model_group_id = model_group.id if model_group is not None else None
        if model_group_id is None:
            raise ValueError("Unable to set the active model. Train a model first")
        # Fire a request to the server to set a model from the group as active
        url = f"{self.base_url}/{model_group_id}:activate"
        _ = self.session.get_rest_response(url=url, method="POST")
        logging.info(f"{algorithm_name} model set as active successfully")

    def get_active_model_for_task(self, task: Task) -> Optional[Model]:
        """
        Return the Model details for the currently active model, for a task if any.
        If the task does not have any trained models, this method returns None

        :param task: Task object containing details of the task to get the model for
        :return: Model object representing the currently active model in the
            Intel® Geti™ project, if any
        """
        model_groups = self.get_all_model_groups()
        model_id: Optional[str] = None
        group_id: Optional[str] = None
        for group in model_groups:
            if not group.has_trained_models:
                continue
            if group.algorithm.task_type != task.type:
                continue
            model_summary = group.get_latest_model()
            if model_summary is not None:
                if model_summary.active_model:
                    model_id = model_summary.id
                    group_id = group.id
                    break
        if model_id is not None:
            return self._get_model_detail(group_id=group_id, model_id=model_id)
        else:
            # Sometimes the `active_model` flag is not set by the server, even though
            # there is a model available for the task. In that case we fall back to
            # returning the latest model in the available model group
            if len(model_groups) == 1:
                model_summary_no_active_check = model_groups[0].get_latest_model()
                if model_summary_no_active_check is not None:
                    model_id = model_summary_no_active_check.id
                    group_id = model_groups[0].id
                    if model_id is not None:
                        return self._get_model_detail(
                            group_id=group_id, model_id=model_id
                        )
        return None

    def _download_model(self, model: ModelType, path_to_folder: str) -> ModelType:
        """
        Download a Model or OptimizedModel.

        :param model: Model or OptimizedModel to download
        :return: Model or OptimizedModel object holding the details of the downloaded
            model
        """
        if isinstance(model, Model):
            url = f"{self.base_url}/{model.model_group_id}/models/{model.id}/export"
            filename = f"{model.name}_base.zip"
        elif isinstance(model, OptimizedModel):
            url = (
                f"{self.base_url}/{model.model_group_id}/models/"
                f"{model.previous_trained_revision_id}/optimized_models/"
                f"{model.id}/export"
            )
            filename = f"{model.name}_{model.optimization_type}_optimized.zip"
        else:
            raise ValueError(
                f"Invalid model type: `{type(model)}. Unable to download model data."
            )
        response = self.session.get_rest_response(
            url=url, method="GET", contenttype="zip"
        )
        model_folder = os.path.join(path_to_folder, "models")
        os.makedirs(model_folder, exist_ok=True, mode=0o770)
        model_filepath = os.path.join(model_folder, filename)
        with open(model_filepath, "wb") as f:
            f.write(response.content)
        return model

    def download_active_model_for_task(
        self, path_to_folder: str, task: Task
    ) -> Optional[Model]:
        """
        Download the currently active model for the task.
        If the task does not have an active model yet, this method returns None

        This method will create a directory 'models' in the path specified in
        `path_to_folder`

        :param path_to_folder: Path to the target folder in which to save the active
            model, and all optimized models derived from it.
        :param task: Task object containing details of the task to download the model
            for
        :return: Model instance holding the details of the active model
        """
        model = self.get_active_model_for_task(task=task)
        if model is None:
            logging.info(
                f"Project '{self.project.name} does not have any trained models yet, "
                f"unable to download active model."
            )
            return None
        model_filepath = os.path.join(path_to_folder, "models")
        logging.info(
            f"Downloading active model for task {task.title} in project "
            f"{self.project.name} to folder {model_filepath}..."
        )
        self._download_model(model, path_to_folder=path_to_folder)
        for optimized_model in model.optimized_models:
            self._download_model(optimized_model, path_to_folder=path_to_folder)
        model_info_filepath = os.path.join(
            model_filepath, f"{task.type}_model_details.json"
        )
        with open(model_info_filepath, "w") as f:
            json.dump(model.to_dict(), f, indent=4)
        return model

    def get_all_active_models(self) -> List[Optional[Model]]:
        """
        Return the Model details for the active model for all tasks in the project,
        if the tasks have any.

        This method returns a list of Models, where the index of the Model in the list
        corresponds to the index of the task in list of trainable tasks for the project.

        If any of the tasks do not have a trained model, the entry corresponding to
        the index of that task will be None

        :return: Model object representing the currently active model for the task in
            the Intel® Geti™ project, if any
        """
        return [
            self.get_active_model_for_task(task=task)
            for task in self.project.get_trainable_tasks()
        ]

    def download_all_active_models(self, path_to_folder: str) -> List[Optional[Model]]:
        """
        Download the active models for all tasks in the project.

        This method will create a directory 'models' in the path specified in
        `path_to_folder`

        :param path_to_folder: Path to the target folder in which to save the active
            models, and all optimized models derived from them.
        :return: List of Model objects representing the currently active models
            (if any) for all tasks in the Intel® Geti™ project. The index of the
            Model in the list corresponds to the index of the task in the list of
            trainable tasks for the project.
        """
        return [
            self.download_active_model_for_task(
                path_to_folder=path_to_folder, task=task
            )
            for task in self.project.get_trainable_tasks()
        ]

    def get_model_for_job(self, job: Job, check_status: bool = True) -> Model:
        """
        Return the model that was created by the `job` from the Intel® Geti™ server.

        :param job: Job to retrieve the model for
        :param check_status: True to first update the status of the job, to make sure
            it is finished. Setting this to False will not update the job status.
        :return: Model produced by the job
        """
        if check_status:
            job.update(self.session)
        if self.session.version.is_sc_mvp or self.session.version.is_sc_1_1:
            job_pid = job.project_id
        else:
            job_pid = job.metadata.project.id
        if job_pid != self.project.id:
            raise ValueError(
                f"Cannot get model for job `{job.description}`. This job does not "
                f"belong to the project managed by this ModelClient instance."
            )
        if job.state != JobState.FINISHED:
            raise ValueError(
                f"Job `{job.description}` is not finished yet, unable to retrieve "
                f"model for the job. Please wait until job is finished"
            )
        metadata = job.metadata
        task_data = metadata.task
        version = task_data.model_version
        algorithm = self.supported_algos.get_by_model_template(
            task_data.model_template_id
        )
        if hasattr(task_data, "name") and job.type in (
            JobType.TRAIN,
            JobType.INFERENCE,
            JobType.EVALUATE,
        ):
            task_name = task_data.name
            task = next(
                (
                    task
                    for task in self.project.get_trainable_tasks()
                    if task.title == task_name
                )
            )
            model = self.get_model_by_algorithm_task_and_version(
                algorithm=algorithm, version=version, task=task
            )
            return model
        if job.type == JobType.OPTIMIZATION:
            model_group_id, optimized_model_id = None, None
            if hasattr(metadata, "model_group_id"):
                model_group_id = metadata.model_group_id
            if hasattr(metadata, "optimized_model_id"):
                optimized_model_id = metadata.optimized_model_id
            if model_group_id is not None and optimized_model_id is not None:
                return self._get_model_detail(model_group_id, optimized_model_id)
        raise ValueError(
            f"Unable to retrieve model for job {job.name} of type {job.type}. Getting "
            f"the model for this job type is not supported. "
        )

    def get_task_for_model(self, model: Union[Model, OptimizedModel]) -> Task:
        """
        Return the task to which a certain model belongs, if possible. This method only
        works when the model identifiers are still in place, if they have been stripped
        it will raise a ValueError.

        If the model does not match any task in the project, this method will raise an
        error.

        :param model: Model or OptimizedModel to find the task for
        :return: Task for which the model was trained
        """
        project_model_groups = self.get_all_model_groups()
        tasks = self.project.get_trainable_tasks()
        model_group: Optional[ModelGroup] = None
        error_msg = (
            f"Unable to match model '{model}' to any task in project "
            f"{self.project}. "
        )
        if model.model_group_id is None:
            raise ValueError(
                error_msg + "The model does not contain a model group identifier"
            )
        for group in project_model_groups:
            if group.id == model.model_group_id:
                model_group = group
                break
        if model_group is None:
            raise ValueError(
                error_msg + "The model does not belong to any of the model groups in "
                "the project."
            )
        else:
            task_id = model_group.task_id
            model_task: Optional[Task] = None
            for task in tasks:
                if task.id == task_id:
                    model_task = task
                    break
        if model_task is None:
            raise ValueError(
                error_msg + f"Model was found on the server but could not be linked "
                f"to specific task for 'task_id={model_group.task_id}'."
            )
        else:
            return model_task

    def optimize_model(self, model: Model, optimization_type: str = "pot") -> Job:
        """
        Start an optimization job for the specified `model`.

        :param model: Model to optimize
        :param optimization_type: Type of optimization to run. Currently supported
            values: ["pot", "nncf"]. Case insensitive. Defaults to "pot"
        :return: Job object referring to the optimization job running on the
            Intel® Geti™ server.
        """
        if isinstance(model, OptimizedModel):
            raise ValueError(
                f"Model {model.name} is already optimized, please specify a base "
                f"model for optimization instead."
            )
        valid_optimization_types = ["pot", "nncf"]
        optimization_type = optimization_type.lower()
        if optimization_type not in valid_optimization_types:
            raise ValueError(
                f"Invalid optimization type specified: `{optimization_type}`. Valid "
                f"options are: {valid_optimization_types}"
            )
        optimize_model_url = model.base_url + "/optimize"
        payload = {
            "enable_nncf_optimization": optimization_type == "nncf",
            "enable_pot_optimization": optimization_type == "pot",
        }
        response = self.session.get_rest_response(
            url=optimize_model_url, method="POST", data=payload
        )
        job = get_job_with_timeout(
            job_id=response["job_id"],
            session=self.session,
            workspace_id=self.workspace_id,
            job_type="optimization",
        )
        return job

    def monitor_job(self, job: Job, timeout: int = 10000, interval: int = 15) -> Job:
        """
        Monitor and print the progress of a `job`. Program execution is
        halted until the job has either finished, failed or was cancelled.

        Progress will be reported in 15s intervals

        :param job: job to monitor
        :param timeout: Timeout (in seconds) after which to stop the monitoring
        :param interval: Time interval (in seconds) at which the ModelClient polls
            the server to update the status of the jobs. Defaults to 15 seconds
        :return: job with it's status updated
        """
        return monitor_job(
            session=self.session, job=job, timeout=timeout, interval=interval
        )



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Dict, Optional

from geti_sdk.data_models.containers import AlgorithmList
from geti_sdk.data_models.enums import TaskType
from geti_sdk.data_models.project import Project
from geti_sdk.http_session import GetiSession


def get_supported_algorithms(
    rest_session: GetiSession,
    task_type: Optional[TaskType] = None,
    project: Optional[Project] = None,
    workspace_id: Optional[str] = None,
) -> AlgorithmList:
    """
    Return the list of supported algorithms (including algorithm metadata) for the
    cluster.

    :param rest_session: HTTP session to the cluster
    :param task_type: Optional TaskType for which to get the supported algorithms.
    :param project: Project to get the supported algorithms for. NOTE: `project` is
        not required for Geti versions v1.8 and lower, but is mandatory for v1.9 and up.
    :param workspace_id: ID of the workspace in which the project to get the supported
        algorithms for lives. NOTE: This is not required for Geti versions v1.8 and
        lower, but is mandatory for v1.9 and up.
    :return: AlgorithmList holding the supported algorithms
    """
    if (workspace_id is None) or (project is None):
        raise ValueError(
            "For Geti v1.9 or higher, passing `workspace_id` and `project` is "
            "mandatory in order to retrieve the supported algorithms"
        )
    url = f"workspaces/{workspace_id}/projects/{project.id}/supported_algorithms"

    algorithm_rest_response = rest_session.get_rest_response(url=url, method="GET")

    if task_type:
        filtered_response = [
            algo
            for algo in algorithm_rest_response["supported_algorithms"]
            if algo["task_type"].upper() == task_type.name
        ]
        algorithm_rest_response["items"] = filtered_response
    return AlgorithmList.from_rest(algorithm_rest_response)


def get_default_algorithm_info(
    session: GetiSession,
    workspace_id: str,
    project: Project,
) -> Dict[TaskType, str]:
    """
    Return the names of the default algorithms for the tasks in the `project`. The
    returned response is a map of TaskType to the default algorithm name for that task

    :param session: GetiSession to the Intel Geti server
    :param workspace_id: Workspace ID in which the project to retrieve the default
        algorithms for lives.
    :param project: Project to retrieve the default algorithms for
    :return: Dictionary mapping the default algorithm name to the TaskType, for each
        task in the project
    """
    if (workspace_id is None) or (project is None):
        raise ValueError(
            "For Geti v1.9 or higher, passing `workspace_id` and `project` is "
            "mandatory in order to retrieve the supported algorithms"
        )
    url = f"workspaces/{workspace_id}/projects/{project.id}/supported_algorithms"

    algorithm_rest_response = session.get_rest_response(url=url, method="GET")
    defaults = algorithm_rest_response.get("default_algorithms", None)
    if defaults is None:
        raise ValueError(
            "The `supported_algorithms` did not return a response to obtain the "
            "default algorithms. Most likely the Geti server you are using does not "
            "support this functionality yet."
        )
    task_type_names = [
        task.type.value.lower() for task in project.get_trainable_tasks()
    ]
    result: Dict[TaskType, str] = {}
    for entry in defaults:
        task_type = entry["task_type"]
        if task_type in task_type_names:
            result.update({TaskType(task_type): entry["model_template_id"]})
    return result



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# noqa: D104

from .annotation_kind import AnnotationKind
from .annotation_state import AnnotationState
from .configuration_enums import ConfigurationEntityType
from .deployment_state import DeploymentState
from .domain import Domain
from .job_state import JobState
from .job_type import JobType
from .media_type import MediaType
from .model_status import ModelStatus
from .optimization_type import OptimizationType
from .prediction_mode import PredictionMode
from .shape_type import ShapeType
from .task_type import TaskType

__all__ = [
    "TaskType",
    "MediaType",
    "ShapeType",
    "AnnotationKind",
    "AnnotationState",
    "PredictionMode",
    "ConfigurationEntityType",
    "Domain",
    "ModelStatus",
    "OptimizationType",
    "JobType",
    "JobState",
    "DeploymentState",
]



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import os
from abc import abstractmethod
from glob import glob
from typing import Dict, List, Optional, Union

from geti_sdk.data_models.annotations import Annotation
from geti_sdk.data_models.enums import TaskType
from geti_sdk.data_models.media import MediaInformation


class AnnotationReader:
    """
    Base class for annotation reading, to handle loading and converting annotations
    to Sonoma Creek format
    """

    def __init__(
        self,
        base_data_folder: str,
        annotation_format: str = ".json",
        task_type: Union[TaskType, str] = TaskType.DETECTION,
    ):
        if task_type is not None and not isinstance(task_type, TaskType):
            task_type = TaskType(task_type)
        self.base_folder = base_data_folder
        self.annotation_format = annotation_format
        self.task_type = task_type

        self._filepaths: Optional[List[str]] = None

    @abstractmethod
    def get_data(
        self,
        filename: str,
        label_name_to_id_mapping: dict,
        media_information: MediaInformation,
        preserve_shape_for_global_labels: bool = False,
    ) -> List[Annotation]:
        """
        Get annotation data for a certain filename
        """
        raise NotImplementedError

    def get_data_filenames(self) -> List[str]:
        """
        Return a list of annotation files found in the `base_data_folder`.

        :return: List of filenames (excluding extension) for all annotation files in
            the data folder
        """
        if self._filepaths is None:
            filepaths = glob(
                os.path.join(self.base_folder, f"*{self.annotation_format}")
            )
            self._filepaths = [
                os.path.splitext(os.path.basename(filepath))[0]
                for filepath in filepaths
            ]
        return self._filepaths

    @abstractmethod
    def get_all_label_names(self) -> List[str]:
        """
        Return a list of unique label names that were found in the annotation data
        folder belonging to this AnnotationReader instance.
        """
        raise NotImplementedError

    def prepare_and_set_dataset(
        self,
        task_type: Union[TaskType, str],
        previous_task_type: Optional[TaskType] = None,
    ) -> None:
        """
        Prepare a dataset for uploading annotations for a certain task_type.

        :param task_type: TaskType to prepare the dataset for
        :param previous_task_type: Optional type of the (trainable) task preceding
            the current task in the pipeline. This is only used for global tasks
        """
        if not isinstance(task_type, TaskType):
            task_type = TaskType(task_type)
        if task_type in [
            TaskType.DETECTION,
            TaskType.SEGMENTATION,
            TaskType.CLASSIFICATION,
        ]:
            self.task_type = task_type
        else:
            raise ValueError(f"Unsupported task_type {task_type}")

    @property
    def applied_filters(self) -> List[Dict[str, Union[List[str], str]]]:
        """
        Return a list of dictionaries representing the filter settings that have
        been applied to the dataset, if any.

        Dictionaries in this list contain two keys:

        - 'labels'      -- List of label names which has been filtered on
        - 'criterion'   -- String representing the criterion that has been used in the
                           filtering. Can be 'OR', 'AND', 'XOR' or 'NOT'.

        :return: List of filter settings that have been applied to the dataset. Returns
            an empty list if no filters have been applied.
        """
        return []



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
from typing import BinaryIO, Dict, Optional, Union


class GetiRequestException(Exception):
    """
    Exception representing an unsuccessful http request to the Intel® Geti™ server.
    """

    def __init__(
        self,
        method: str,
        url: str,
        status_code: int,
        request_data: Dict[str, Union[dict, str, list, BinaryIO]],
        response_data: Optional[Union[dict, str, list]] = None,
    ):
        """
        Raise this exception upon unsuccessful requests to the Intel® Geti™ server.

        :param method: Method that was used for the request, e.g. 'POST' or 'GET', etc.
        :param url: URL to which the request was made
        :param status_code: HTTP status code returned for the request
        :param request_data: Data that was included with the request.
        :param response_data: Optional data that was returned in response to the
            request, if any
        """
        self.method = method
        self.url = url
        self.status_code = status_code
        self.request_data = request_data

        self.response_message: Optional[str] = None
        self.response_error_code: Optional[str] = None

        if response_data:
            if isinstance(response_data, dict):
                self.response_message = response_data.get("message", None)
                self.response_error_code = response_data.get("error_code", None)
            elif isinstance(response_data, str):
                self.response_message = response_data
                self.response_error_code = self.status_code
            else:
                logging.warning(
                    f"Received unexpected response type `{type(response_data)}` from "
                    f"server. Response: {response_data}"
                )

    def __str__(self) -> str:
        """
        Return string representation of the unsuccessful http request to the
        Intel® Geti™ server.
        """
        error_str = (
            f"{self.method} request to '{self.url}' failed with status code "
            f"{self.status_code}."
        )
        if self.response_error_code and self.response_message:
            error_str += (
                f" Server returned error code '{self.response_error_code}' "
                f"with message '{self.response_message}'"
            )
        return error_str



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
from pprint import pformat
from typing import Any, Dict, List, Optional, Tuple

import attr

from geti_sdk.data_models.enums import JobState, JobType
from geti_sdk.data_models.project import Dataset
from geti_sdk.data_models.status import StatusSummary
from geti_sdk.data_models.utils import (
    attr_value_serializer,
    str_to_datetime,
    str_to_enum_converter,
    str_to_optional_enum_converter,
)
from geti_sdk.http_session import GetiRequestException, GetiSession
from geti_sdk.platform_versions import GetiVersion


@attr.define(slots=False)
class JobStatus(StatusSummary):
    """
    Current status of a job on the Intel® Geti™ server.

    :var state: Current state of the job
    """

    state: str = attr.field(converter=str_to_enum_converter(JobState), kw_only=True)

    @classmethod
    def from_dict(cls, status_dict: Dict[str, Any]) -> "JobStatus":
        """
        Create a JobStatus object from a dictionary.

        :param status_dict: Dictionary representing a status, as returned by the
            Intel® Geti™ /status and /jobs endpoints
        :return: JobStatus object holding the status data contained in `status_dict`
        """
        return cls(**status_dict)


@attr.define
class TaskMetadata:
    """
    Metadata related to a task on the Intel® Geti™ cluster.

    :var name: Name of the task
    :var model_template_id: Identifier of the model template used by the task
    :var model_architecture: Name of the neural network architecture used for the model
    :var model_version: Version of the model currently used by the job
    :var dataset_storage_id: Unique database ID of the dataset storage used by the job
    :var task_id: ID of the task to which the TaskStatus object applies. Only used in
        Geti v1.1 and up
    """

    model_architecture: Optional[str] = None
    model_template_id: Optional[str] = None
    model_version: Optional[int] = None
    name: Optional[str] = None
    dataset_storage_id: Optional[str] = None
    task_id: Optional[str] = None  # Added in Geti v1.1


@attr.define
class TestMetadata:
    """
    Metadata related to a model test job on the GETi cluster.

    :var model_template_id: Identifier of the model template used in the test
    :var model_architecture: Name of the neural network architecture used for the model
    :var datasets: List of dictionaries, each dictionary holding the id and name of a
        dataset used in the test
    """

    model_architecture: str
    model_template_id: str
    datasets: List[Dataset]
    model: Optional[dict] = None  # Added in Geti v1.7


@attr.define
class ProjectMetadata:
    """
    Metadata related to a project on the GETi cluster.

    :var name: Name of the project
    :var id: ID of the project
    """

    name: Optional[str] = None
    id: Optional[str] = None


@attr.define
class ModelMetadata:
    """
    Metadata for a Job related to a model on the GETi cluster.

    :var model_storage_id: ID of the model storage in which the model lives
    :var model_id: ID of the model
    :var model_activated: True if the model has been activated succesfully
        This is applicable after a training job. Defaults to `None`
    """

    model_storage_id: str
    model_id: str
    model_activated: Optional[bool] = None  # Added in Geti v1.14


@attr.define
class ScoreMetadata:
    """
    Metadata element containing scores for the tasks in the project

    :var task_id: ID of the task for which the score was achieved
    :var score: Performance score for the model for the task
    """

    task_id: str
    score: float


@attr.define
class JobMetadata:
    """
    Metadata for a particular job on the GETi cluster.

    :var task: TaskMetadata object holding information regarding the task from which
        the job originates
    :var base_model_id: Optional unique database ID of the base model. Only used for
        optimization jobs
    :var model_storage_id: Optional unique database ID of the model storage used by
        the job.
    :var optimization_type: Optional type of the optimization method used in the job.
        Only used for optimization jobs
    :var optimized_model_id: Optional unique database ID of the optimized model
        produced by the job. Only used for optimization jobs.
    :var scores: List of scores for the job. Added in Geti v1.1
    """

    task: Optional[TaskMetadata] = None
    project: Optional[ProjectMetadata] = None
    test: Optional[TestMetadata] = None
    base_model_id: Optional[str] = None
    model_storage_id: Optional[str] = None
    optimization_type: Optional[str] = None
    optimized_model_id: Optional[str] = None
    scores: Optional[List[ScoreMetadata]] = None
    trained_model: Optional[ModelMetadata] = None  # Added in Geti v1.7
    warnings: Optional[List[dict]] = None  # Added in Geti v1.13 for dataset import jobs
    supported_project_types: Optional[List[dict]] = (
        None  # Added in Geti v1.13 for dataset import jobs
    )
    project_id: Optional[str] = None  # Added in Geti v1.13 for dataset import jobs


@attr.define
class JobCancellationInfo:
    """
    Information relating to the cancellation of a Job in Intel Geti

    :var is_cancelled: True if the job is cancelled, False otherwise
    :var user_uid: Unique ID of the User who cancelled the Job
    :var cancel_time: Time at which the Job was cancelled
    """

    is_cancelled: bool = False
    user_uid: Optional[str] = None
    cancel_time: Optional[str] = attr.field(converter=str_to_datetime, default=None)


@attr.define(slots=False)
class Job:
    """
    Representation of a job running on the GETi cluster.

    :var name: Name of the job
    :var id: Unique database ID of the job
    :var description: Brief description of the job
    :var type: Type of the job
    :var metadata: JobMetadata object holding metadata for the job
    """

    name: str
    id: str
    type: str = attr.field(converter=str_to_enum_converter(JobType))
    metadata: JobMetadata
    description: Optional[str] = None
    creation_time: Optional[str] = attr.field(converter=str_to_datetime, default=None)
    start_time: Optional[str] = attr.field(
        converter=str_to_datetime, default=None
    )  # Added in Geti v1.7
    end_time: Optional[str] = attr.field(
        converter=str_to_datetime, default=None
    )  # Added in Geti v1.7
    author: Optional[str] = None  # Added in Geti v1.7
    cancellation_info: Optional[JobCancellationInfo] = None  # Added in Geti v1.7
    state: Optional[str] = attr.field(
        converter=str_to_optional_enum_converter(JobState), default=None
    )  # Added in Geti v1.7
    steps: Optional[List[dict]] = None  # Added in Geti v1.7

    def __attrs_post_init__(self):
        """
        Initialize private attributes.
        """
        self._workspace_id: Optional[str] = None
        self._geti_version: Optional[GetiVersion]

    @property
    def workspace_id(self) -> str:
        """
        Return the unique database ID of the workspace to which the job belongs.

        :return: Unique database ID of the workspace to which the job belongs
        """
        if self._workspace_id is None:
            raise ValueError(
                f"Workspace ID for job {self} is unknown, it was never set."
            )
        return self._workspace_id

    @workspace_id.setter
    def workspace_id(self, workspace_id: str):
        """
        Set the workspace id for the job.

        :param workspace_id: Unique database ID of the workspace to which the job
            belongs
        """
        self._workspace_id = workspace_id

    @property
    def relative_url(self) -> str:
        """
        Return the url at which the Job can be addressed on the GETi cluster, relative
        to the url of the cluster itself.

        :return: Relative url for the Job instance
        """
        return f"workspaces/{self.workspace_id}/jobs/{self.id}"

    def update(self, session: GetiSession) -> "Job":
        """
        Update the job status to its current value, by making a request to the GETi
        cluster addressed by `session`.

        :param session: GetiSession to the cluster from which the Job originates
        :raises ValueError: If no workspace_id has been set for the job prior to
            calling this method
        :return: Job with its status updated
        """
        try:
            response = session.get_rest_response(url=self.relative_url, method="GET")
        except GetiRequestException as job_error:
            if job_error.status_code == 403:
                raise GetiRequestException(
                    method=job_error.method,
                    url=job_error.url,
                    status_code=404,
                    request_data=job_error.request_data,
                    response_data={
                        "message": f"Job with id {self.id} does not exist on the platform",
                        "error_code": "job_not_found",
                    },
                )
            else:
                raise job_error

        self.steps = response.get("steps", None)
        self.state = JobState(response["state"])

        if self._geti_version is None:
            self.geti_version = session.version
        return self

    def cancel(self, session: GetiSession) -> "Job":
        """
        Cancel and delete the job, by making a request to the GETi cluster addressed
        by `session`.

        :param session: GetiSession to the cluster on which the Job is running
        :return: Job with updated status
        """
        try:
            session.get_rest_response(
                url=self.relative_url, method="DELETE", allow_text_response=True
            )
            self.state = JobState.CANCELLED
        except GetiRequestException as error:
            if error.status_code == 404:
                logging.info(
                    f"Job '{self.name}' is not active anymore, unable to delete."
                )
                self.state = JobState.INACTIVE
            else:
                raise error
        return self

    @property
    def overview(self) -> str:
        """
        Return a string that shows an overview of the job.

        :return: String holding an overview of the job
        """
        return pformat(self.to_dict())

    def to_dict(self) -> Dict[str, Any]:
        """
        Return the dictionary representation of the job.

        :return:
        """
        return attr.asdict(self, recurse=True, value_serializer=attr_value_serializer)

    @property
    def is_finished(self) -> bool:
        """
        Return True if the job finished successfully, False otherwise
        """
        return self.state == JobState.FINISHED

    @property
    def is_running(self) -> bool:
        """
        Return True if the job is currently running, False otherwise
        """
        return self.state == JobState.RUNNING

    def _get_step_information(self) -> Tuple[int, int]:
        """
        Return the current step and the total number of steps in the job
        """
        if self.steps is not None and len(self.steps) != 0:
            # Job is split in steps, valid from Geti v1.10
            total = len(self.steps)
            steps_complete = 0
            for step in self.steps:
                step_state = step.get("state", "waiting")
                if step_state == "finished":
                    steps_complete += 1
            current = steps_complete + 1
        else:
            return 0, 1
        return current, total

    @property
    def total_steps(self) -> int:
        """
        Return the total number of steps in the job
        """
        return self._get_step_information()[1]

    @property
    def current_step(self) -> int:
        """
        Return the current step for the job
        """
        return self._get_step_information()[0]

    @property
    def current_step_message(self) -> str:
        """
        Return the description of the current step for the job

        :return: String containing the current step name/description. If
            for whatever reason the current step cannot be determined,
            an empty string is returned
        """
        current_step_index = self.current_step - 1
        if current_step_index < 0 or current_step_index >= len(self.steps):
            if self.state != JobState.SCHEDULED:
                return ""
            else:
                return "Awaiting job execution"
        return self.steps[current_step_index].get("step_name", "")

    @property
    def current_step_progress(self) -> float:
        """
        Return the progress of the current step for the job

        :return: float indicating the progress of the current step in the job
        """
        current_step_index = self.current_step - 1
        if current_step_index < 0 or current_step_index >= len(self.steps):
            return 0.0
        return self.steps[current_step_index].get("progress", 0.0)

    @property
    def geti_version(self) -> GetiVersion:
        """
        Return the version of the Intel Geti instance from which the job originates.

        :return: Version of the Intel Geti instance from which the job originates
        """
        if self._geti_version is None:
            raise ValueError(
                f"Geti version for job {self} is unknown, it was never set."
            )
        return self._geti_version

    @geti_version.setter
    def geti_version(self, geti_version: GetiVersion):
        """
        Set the version of the Intel Geti instance for the job.

        :param geti_version: Version of the Intel Geti instance from which the job originates
        """
        self._geti_version = geti_version



# INTEL CONFIDENTIAL
#
# Copyright (C) 2023 Intel Corporation
#
# This software and the related documents are Intel copyrighted materials, and
# your use of them is governed by the express license under which they were provided to
# you ("License"). Unless the License provides otherwise, you may not use, modify, copy,
# publish, distribute, disclose or transmit this software or the related documents
# without Intel's prior written permission.
#
# This software and the related documents are provided as is,
# with no express or implied warranties, other than those that are expressly stated
# in the License.

"""
The package contains classes for inference results post-processing and conversion to internal Prediction entities.

The package contains the following classes:
    - `AnomalyToPredictionConverter` - class for converting anomaly classification / segmentation / detection results to internal Prediction entities
    - `ClassificationToPredictionConverter` - class for converting classification results to internal Prediction entities
    - `DetectionToPredictionConverter` - class for converting detection results to internal Prediction entities
    - `MaskToAnnotationConverter` - class for converting rotated detection results to internal Prediction entities
    - `RotatedRectToPredictionConverter` - class for converting rotated detection results to internal Prediction entities
    - `SegmentationToPredictionConverter` - class for converting segmentation results to internal Prediction entities

    - `ConverterFactory` - factory class for creating the appropriate converter based on the domain of the inference results
"""

from .results_to_prediction_converter import ConverterFactory

__all__ = ["ConverterFactory"]
